from typing import Callable, Generic, Iterator, Optional, Set, Type

from vellum.workflows.context import execution_context, get_parent_context
from vellum.workflows.errors.types import WorkflowError, WorkflowErrorCode
from vellum.workflows.events.workflow import is_workflow_event
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases.base_adornment_node import BaseAdornmentNode
from vellum.workflows.nodes.utils import create_adornment
from vellum.workflows.outputs.base import BaseOutput, BaseOutputs
from vellum.workflows.references.output import OutputReference
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.generics import StateType
from vellum.workflows.workflows.event_filters import all_workflow_event_filter


class TryNode(BaseAdornmentNode[StateType], Generic[StateType]):
    """
    Used to execute a Subworkflow and handle errors.

    on_error_code: Optional[WorkflowErrorCode] = None - The error code to handle
    subworkflow: Type["BaseWorkflow"] - The Subworkflow to execute
    """

    on_error_code: Optional[WorkflowErrorCode] = None

    class Outputs(BaseAdornmentNode.Outputs):
        error: Optional[WorkflowError] = None

    def run(self) -> Iterator[BaseOutput]:
        parent_context = get_parent_context()
        with execution_context(parent_context=parent_context):
            subworkflow = self.subworkflow(
                parent_state=self.state,
                context=WorkflowContext.create_from(self._context),
            )
            subworkflow_stream = subworkflow.stream(
                event_filter=all_workflow_event_filter,
                node_output_mocks=self._context._get_all_node_output_mocks(),
            )

        outputs: Optional[BaseOutputs] = None
        exception: Optional[NodeException] = None
        fulfilled_output_names: Set[str] = set()

        for event in subworkflow_stream:
            self._context._emit_subworkflow_event(event)
            if exception:
                continue

            if not is_workflow_event(event):
                continue

            if event.workflow_definition != self.subworkflow:
                continue

            if event.name == "workflow.execution.streaming":
                if event.output.is_fulfilled:
                    fulfilled_output_names.add(event.output.name)
                yield event.output
            elif event.name == "workflow.execution.fulfilled":
                outputs = event.outputs
            elif event.name == "workflow.execution.paused":
                exception = NodeException(
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                    message="Subworkflow unexpectedly paused within Try Node",
                )
            elif event.name == "workflow.execution.rejected":
                if self.on_error_code and self.on_error_code != event.error.code:
                    exception = NodeException(
                        code=WorkflowErrorCode.INVALID_OUTPUTS,
                        message=f"""Unexpected rejection: {event.error.code.value}.
Message: {event.error.message}""",
                    )
                else:
                    outputs = self.Outputs(error=event.error)

        if exception:
            raise exception

        if outputs is None:
            raise NodeException(
                code=WorkflowErrorCode.INVALID_OUTPUTS,
                message="Expected to receive outputs from Try Node's subworkflow",
            )

        # For any outputs somehow in our final fulfilled outputs array,
        # but not fulfilled by the stream.
        for descriptor, value in outputs:
            if descriptor.name not in fulfilled_output_names:
                yield BaseOutput(
                    name=descriptor.name,
                    value=value,
                )

    @classmethod
    def wrap(cls, on_error_code: Optional[WorkflowErrorCode] = None) -> Callable[..., Type["TryNode"]]:
        return create_adornment(cls, attributes={"on_error_code": on_error_code})

    @classmethod
    def __annotate_outputs_class__(cls, outputs_class: Type[BaseOutputs], reference: OutputReference) -> None:
        if reference.name == "error":
            raise ValueError("`error` is a reserved name for TryNode.Outputs")

        super().__annotate_outputs_class__(outputs_class, reference)



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import DeploymentRead
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_text_prompt_deployment.workflow import BasicTextPromptDeployment
from tests.workflows.basic_text_prompt_deployment.workflow_with_prompt_deployment_json_reference import (
    WorkflowWithPromptDeploymentJsonReferenceWorkflow,
)


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment = DeploymentRead(
        id=str(uuid4()),
        created=datetime.now(),
        label="Example Prompt Deployment",
        name="example_prompt_deployment",
        last_deployed_on=datetime.now(),
        input_variables=[],
        active_model_version_ids=[],
        last_deployed_history_item_id=str(uuid4()),
    )
    vellum_client.deployments.retrieve.return_value = deployment

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicTextPromptDeployment)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "52995b50-84c9-465f-8a4b-a4ee2a92e388",
                "key": "city",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
            {
                "id": "aa3ca842-250c-4a3f-853f-23928c28d0f8",
                "key": "date",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [
        {
            "id": "a609ab19-db1b-4cd0-bdb0-aee5ed31dc28",
            "key": "text",
            "type": "STRING",
        }
    ]

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 2
    assert len(workflow_raw_data["nodes"]) == 3

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "d680afbd-de64-4cf6-aa50-912686c48c64",
        "type": "ENTRYPOINT",
        "inputs": [],
        "base": None,
        "definition": None,
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "7065a943-1cab-4afd-9690-e678c5b74a2f",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    prompt_node = workflow_raw_data["nodes"][1]
    assert prompt_node == {
        "id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
        "type": "PROMPT",
        "inputs": [
            {
                "id": "947d7ead-0fad-4e5f-aa3a-d06029ac94bc",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "52995b50-84c9-465f-8a4b-a4ee2a92e388"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "3deebdd7-2900-4d8c-93f2-e5b90649ac42",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "aa3ca842-250c-4a3f-853f-23928c28d0f8"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Prompt Deployment Node",
            "output_id": "4d38b850-79e3-4b85-9158-a41d0c535410",
            "error_output_id": None,
            "array_output_id": "0cf47d33-6d5f-466f-b826-e814f1d0348b",
            "source_handle_id": "2f26c7e0-283d-4f04-b639-adebb56bc679",
            "target_handle_id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "variant": "DEPLOYMENT",
            "prompt_deployment_id": deployment.id,
            "release_tag": "LATEST",
            "ml_model_fallbacks": ["gpt-4o", "gemini-1.5-pro"],
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "PromptDeploymentNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "prompt_deployment_node", "node"],
        },
        "definition": {
            "name": "ExamplePromptDeploymentNode",
            "module": ["tests", "workflows", "basic_text_prompt_deployment", "workflow"],
        },
        "trigger": {
            "id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "2f26c7e0-283d-4f04-b639-adebb56bc679", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {"id": "180355a8-e67c-4ce6-9ac3-e5dbb75a6629", "name": "json", "type": "JSON", "value": None},
            {"id": "4d38b850-79e3-4b85-9158-a41d0c535410", "name": "text", "type": "STRING", "value": None},
            {"id": "0cf47d33-6d5f-466f-b826-e814f1d0348b", "name": "results", "type": "ARRAY", "value": None},
        ],
    }

    final_output_node = workflow_raw_data["nodes"][2]
    assert final_output_node == {
        "id": "64ff72c7-8ffc-4e1f-b7a7-e7cd0697f576",
        "type": "TERMINAL",
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "final_output_node",
                "node",
            ],
            "name": "FinalOutputNode",
        },
        "definition": None,
        "inputs": [
            {
                "id": "78aeb65b-3491-4d2a-8c47-401d4cb3d560",
                "key": "node_input",
                "value": {
                    "combinator": "OR",
                    "rules": [
                        {
                            "data": {
                                "node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
                                "output_id": "4d38b850-79e3-4b85-9158-a41d0c535410",
                            },
                            "type": "NODE_OUTPUT",
                        }
                    ],
                },
            }
        ],
        "data": {
            "label": "Final Output",
            "name": "text",
            "node_input_id": "78aeb65b-3491-4d2a-8c47-401d4cb3d560",
            "output_id": "a609ab19-db1b-4cd0-bdb0-aee5ed31dc28",
            "output_type": "STRING",
            "target_handle_id": "dced939a-9122-4290-8482-7daa9525dad6",
        },
        "display_data": {
            "position": {
                "x": 400.0,
                "y": -50.0,
            },
        },
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "8961d02b-074e-45ab-9f77-4e94606a4344",
            "source_handle_id": "7065a943-1cab-4afd-9690-e678c5b74a2f",
            "source_node_id": "d680afbd-de64-4cf6-aa50-912686c48c64",
            "target_handle_id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "target_node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
            "type": "DEFAULT",
        },
        {
            "id": "c2cbf6ef-8582-45c8-a643-fc6ae8fe482f",
            "source_handle_id": "2f26c7e0-283d-4f04-b639-adebb56bc679",
            "source_node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
            "target_handle_id": "dced939a-9122-4290-8482-7daa9525dad6",
            "target_node_id": "64ff72c7-8ffc-4e1f-b7a7-e7cd0697f576",
            "type": "DEFAULT",
        },
    ]

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicTextPromptDeployment",
        "module": [
            "tests",
            "workflows",
            "basic_text_prompt_deployment",
            "workflow",
        ],
    }


def test_serialize_workflow_with_prompt_and_templating(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment = DeploymentRead(
        id=str(uuid4()),
        created=datetime.now(),
        label="JSON Prompt Deployment",
        name="json_prompt_deployment",
        last_deployed_on=datetime.now(),
        input_variables=[],
        active_model_version_ids=[],
        last_deployed_history_item_id=str(uuid4()),
    )
    vellum_client.deployments.retrieve.return_value = deployment

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=WorkflowWithPromptDeploymentJsonReferenceWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "ad577aca-0797-4deb-a484-574ef1a1f0c7",
                "key": "city",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            },
            {
                "id": "066124c4-42bd-4764-aa75-6f230dbbed4a",
                "key": "date",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [{"id": "a7e4b449-5879-4d0c-8f00-d5d4985eb65c", "key": "text", "type": "STRING"}]

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 3
    assert len(workflow_raw_data["nodes"]) == 4

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "32c7f398-277c-456b-9279-aa1f867fb637",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "cc0f4028-1039-4063-971d-7dacbb01b379"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    prompt_node = workflow_raw_data["nodes"][1]
    assert prompt_node == {
        "id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
        "type": "PROMPT",
        "inputs": [
            {
                "id": "947d7ead-0fad-4e5f-aa3a-d06029ac94bc",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "ad577aca-0797-4deb-a484-574ef1a1f0c7"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "3deebdd7-2900-4d8c-93f2-e5b90649ac42",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "066124c4-42bd-4764-aa75-6f230dbbed4a"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Prompt Deployment Node",
            "output_id": "4d38b850-79e3-4b85-9158-a41d0c535410",
            "error_output_id": None,
            "array_output_id": "0cf47d33-6d5f-466f-b826-e814f1d0348b",
            "source_handle_id": "2f26c7e0-283d-4f04-b639-adebb56bc679",
            "target_handle_id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "variant": "DEPLOYMENT",
            "prompt_deployment_id": deployment.id,
            "release_tag": "LATEST",
            "ml_model_fallbacks": None,
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "PromptDeploymentNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "prompt_deployment_node", "node"],
        },
        "definition": {
            "name": "ExamplePromptDeploymentNode",
            "module": [
                "tests",
                "workflows",
                "basic_text_prompt_deployment",
                "workflow_with_prompt_deployment_json_reference",
            ],
        },
        "trigger": {
            "id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "2f26c7e0-283d-4f04-b639-adebb56bc679", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {"id": "180355a8-e67c-4ce6-9ac3-e5dbb75a6629", "name": "json", "type": "JSON", "value": None},
            {"id": "4d38b850-79e3-4b85-9158-a41d0c535410", "name": "text", "type": "STRING", "value": None},
            {"id": "0cf47d33-6d5f-466f-b826-e814f1d0348b", "name": "results", "type": "ARRAY", "value": None},
        ],
    }

    templating_node = workflow_raw_data["nodes"][2]
    assert templating_node == {
        "id": "51cbe21d-0232-4362-bc54-5bc283297aa6",
        "type": "TEMPLATING",
        "inputs": [
            {
                "id": "7c775379-d589-4d79-b876-dcd224d72966",
                "key": "template",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {
                                "type": "STRING",
                                "value": "The weather in {{ city }} on {{ date }} is {{ weather }}.",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "dec1317a-6900-4858-a5fb-c849254b2c91",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "ad577aca-0797-4deb-a484-574ef1a1f0c7"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "4cc5b9f1-075d-45fd-a978-f530c29c5682",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "066124c4-42bd-4764-aa75-6f230dbbed4a"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "77b352e0-1b77-4d48-9f6f-04ce17fe7286",
                "key": "weather",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
                                "output_id": "180355a8-e67c-4ce6-9ac3-e5dbb75a6629",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Templating Node",
            "output_id": "6834cae4-8173-4fa6-88f7-bc09d335bdd1",
            "error_output_id": None,
            "source_handle_id": "39317827-df43-4f5a-bfbc-20bffc839748",
            "target_handle_id": "58427684-3848-498a-8299-c6b0fc70265d",
            "template_node_input_id": "7c775379-d589-4d79-b876-dcd224d72966",
            "output_type": "STRING",
        },
        "display_data": {"position": {"x": 400.0, "y": -50.0}},
        "base": {
            "name": "TemplatingNode",
            "module": ["vellum", "workflows", "nodes", "core", "templating_node", "node"],
        },
        "definition": {
            "name": "ExampleTemplatingNode",
            "module": [
                "tests",
                "workflows",
                "basic_text_prompt_deployment",
                "workflow_with_prompt_deployment_json_reference",
            ],
        },
        "trigger": {
            "id": "58427684-3848-498a-8299-c6b0fc70265d",
            "merge_behavior": "AWAIT_ATTRIBUTES",
        },
        "ports": [{"id": "39317827-df43-4f5a-bfbc-20bffc839748", "name": "default", "type": "DEFAULT"}],
    }

    final_output_node = workflow_raw_data["nodes"][3]
    assert final_output_node == {
        "id": "53de824d-a41d-4294-b511-c969932b05af",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "text",
            "target_handle_id": "fee3d395-38c3-485f-ab61-1a0fdf71c4ce",
            "output_id": "a7e4b449-5879-4d0c-8f00-d5d4985eb65c",
            "output_type": "STRING",
            "node_input_id": "cf380f81-c5ee-4bc9-8e26-ecf1307733a9",
        },
        "inputs": [
            {
                "id": "cf380f81-c5ee-4bc9-8e26-ecf1307733a9",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "51cbe21d-0232-4362-bc54-5bc283297aa6",
                                "output_id": "6834cae4-8173-4fa6-88f7-bc09d335bdd1",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 600.0, "y": -50.0}},
        "base": {
            "name": "FinalOutputNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
        },
        "definition": None,
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "2c49ffa6-7b9a-49a0-a932-009534556480",
            "source_node_id": "32c7f398-277c-456b-9279-aa1f867fb637",
            "source_handle_id": "cc0f4028-1039-4063-971d-7dacbb01b379",
            "target_node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
            "target_handle_id": "b7605c48-0937-4ecc-914e-0d1058130e65",
            "type": "DEFAULT",
        },
        {
            "id": "a46909ec-9572-43c6-a134-0bd7e2c09f99",
            "source_node_id": "56c74024-19a3-4c0d-a5f5-23e1e9f11b21",
            "source_handle_id": "2f26c7e0-283d-4f04-b639-adebb56bc679",
            "target_node_id": "51cbe21d-0232-4362-bc54-5bc283297aa6",
            "target_handle_id": "58427684-3848-498a-8299-c6b0fc70265d",
            "type": "DEFAULT",
        },
        {
            "id": "1f720900-e5e1-49b7-9910-6ede79f6afd2",
            "source_node_id": "51cbe21d-0232-4362-bc54-5bc283297aa6",
            "source_handle_id": "39317827-df43-4f5a-bfbc-20bffc839748",
            "target_node_id": "53de824d-a41d-4294-b511-c969932b05af",
            "target_handle_id": "fee3d395-38c3-485f-ab61-1a0fdf71c4ce",
            "type": "DEFAULT",
        },
    ]

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "WorkflowWithPromptDeploymentJsonReferenceWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_text_prompt_deployment",
            "workflow_with_prompt_deployment_json_reference",
        ],
    }



# This file was auto-generated by Fern from our API Definition.

import typing

from .named_scenario_input_chat_history_variable_value_request import NamedScenarioInputChatHistoryVariableValueRequest
from .named_scenario_input_json_variable_value_request import NamedScenarioInputJsonVariableValueRequest
from .named_scenario_input_string_variable_value_request import NamedScenarioInputStringVariableValueRequest

NamedScenarioInputRequest = typing.Union[
    NamedScenarioInputStringVariableValueRequest,
    NamedScenarioInputJsonVariableValueRequest,
    NamedScenarioInputChatHistoryVariableValueRequest,
]



import dataclasses
import inspect
from typing import TYPE_CHECKING, Annotated, Any, Callable, Dict, Literal, Optional, Type, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from pydash import snake_case

from vellum import Vellum
from vellum.client.types.function_definition import FunctionDefinition
from vellum.workflows.utils.vellum_variables import vellum_variable_type_to_openapi_type

if TYPE_CHECKING:
    from vellum.workflows.workflows.base import BaseWorkflow

type_map = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "array",
    dict: "object",
    None: "null",
    type(None): "null",
    inspect._empty: "null",
}


def compile_annotation(annotation: Optional[Any], defs: dict[str, Any]) -> dict:
    if annotation is None:
        return {"type": "null"}

    if get_origin(annotation) is Union:
        return {"anyOf": [compile_annotation(a, defs) for a in get_args(annotation)]}

    if get_origin(annotation) is Literal:
        values = list(get_args(annotation))
        types = {type(value) for value in values}
        if len(types) == 1:
            value_type = types.pop()
            if value_type in type_map:
                return {"type": type_map[value_type], "enum": values}
            else:
                return {"enum": values}
        else:
            return {"enum": values}

    if get_origin(annotation) is dict:
        _, value_type = get_args(annotation)
        return {"type": "object", "additionalProperties": compile_annotation(value_type, defs)}

    if get_origin(annotation) is list:
        item_type = get_args(annotation)[0]
        return {"type": "array", "items": compile_annotation(item_type, defs)}

    if get_origin(annotation) is tuple:
        args = get_args(annotation)
        if len(args) == 2 and args[1] is Ellipsis:
            # Tuple[int, ...] with homogeneous items
            return {"type": "array", "items": compile_annotation(args[0], defs)}
        else:
            # Tuple[int, str] with fixed length items
            result = {
                "type": "array",
                "prefixItems": [compile_annotation(arg, defs) for arg in args],
                "minItems": len(args),
                "maxItems": len(args),
            }
            return result

    if dataclasses.is_dataclass(annotation) and isinstance(annotation, type):
        if annotation.__name__ not in defs:
            properties = {}
            required = []
            for field in dataclasses.fields(annotation):
                properties[field.name] = compile_annotation(field.type, defs)
                if field.default is dataclasses.MISSING:
                    required.append(field.name)
                else:
                    properties[field.name]["default"] = _compile_default_value(field.default)
            defs[annotation.__name__] = {"type": "object", "properties": properties, "required": required}
        return {"$ref": f"#/$defs/{annotation.__name__}"}

    if issubclass(annotation, BaseModel):
        if annotation.__name__ not in defs:
            properties = {}
            required = []
            for field_name, field in annotation.model_fields.items():
                # Mypy is incorrect here, the `annotation` attribute is defined on `FieldInfo`
                field_annotation = field.annotation  # type: ignore[attr-defined]
                properties[field_name] = compile_annotation(field_annotation, defs)
                if field.default is PydanticUndefined:
                    required.append(field_name)
                else:
                    properties[field_name]["default"] = _compile_default_value(field.default)
            defs[annotation.__name__] = {"type": "object", "properties": properties, "required": required}

        return {"$ref": f"#/$defs/{annotation.__name__}"}

    return {"type": type_map[annotation]}


def _compile_default_value(default: Any) -> Any:
    if dataclasses.is_dataclass(default):
        return {
            field.name: _compile_default_value(getattr(default, field.name)) for field in dataclasses.fields(default)
        }

    if isinstance(default, BaseModel):
        return {
            field_name: _compile_default_value(getattr(default, field_name))
            for field_name in default.__class__.model_fields.keys()
        }

    return default


def _compile_workflow_deployment_input(input_var: Any) -> dict[str, Any]:
    """
    Converts a deployment workflow input variable to a JSON schema type definition.
    """
    primitive_type = vellum_variable_type_to_openapi_type(input_var.type)
    input_schema = {"type": primitive_type}

    if input_var.default is not None:
        input_schema["default"] = input_var.default.value

    return input_schema


def compile_function_definition(function: Callable) -> FunctionDefinition:
    """
    Converts a Python function into our Vellum-native FunctionDefinition type.
    """

    try:
        signature = inspect.signature(function)
    except ValueError as e:
        raise ValueError(f"Failed to get signature for function {function.__name__}: {str(e)}")

    properties = {}
    required = []
    defs: dict[str, Any] = {}
    for param in signature.parameters.values():
        # Check if parameter uses Annotated type hint
        if get_origin(param.annotation) is Annotated:
            args = get_args(param.annotation)
            actual_type = args[0]
            # Extract description from metadata
            description = args[1] if len(args) > 1 and isinstance(args[1], str) else None

            properties[param.name] = compile_annotation(actual_type, defs)
            if description:
                properties[param.name]["description"] = description
        else:
            properties[param.name] = compile_annotation(param.annotation, defs)

        if param.default is inspect.Parameter.empty:
            required.append(param.name)
        else:
            properties[param.name]["default"] = _compile_default_value(param.default)

    parameters = {"type": "object", "properties": properties, "required": required}
    if defs:
        parameters["$defs"] = defs

    return FunctionDefinition(
        name=function.__name__,
        description=function.__doc__,
        parameters=parameters,
    )


def compile_inline_workflow_function_definition(workflow_class: Type["BaseWorkflow"]) -> FunctionDefinition:
    """
    Converts a base workflow class into our Vellum-native FunctionDefinition type.
    """
    # Get the inputs class for the workflow
    inputs_class = workflow_class.get_inputs_class()
    vars_inputs_class = vars(inputs_class)

    properties = {}
    required = []
    defs: dict[str, Any] = {}

    for name, field_type in inputs_class.__annotations__.items():
        if name.startswith("__"):
            continue

        properties[name] = compile_annotation(field_type, defs)

        # Check if the field has a default value
        if name not in vars_inputs_class:
            required.append(name)
        else:
            # Field has a default value
            properties[name]["default"] = vars_inputs_class[name]

    parameters = {"type": "object", "properties": properties, "required": required}
    if defs:
        parameters["$defs"] = defs

    return FunctionDefinition(
        name=snake_case(workflow_class.__name__),
        description=workflow_class.__doc__,
        parameters=parameters,
    )


def compile_workflow_deployment_function_definition(
    deployment_config: Dict[str, str],
    vellum_client: Vellum,
) -> FunctionDefinition:
    """
    Converts a deployment workflow config into our Vellum-native FunctionDefinition type.

    Args:
        deployment_config: Dict with 'deployment' and 'release_tag' keys
        vellum_client: Vellum client instance
    """
    deployment = deployment_config["deployment"]
    release_tag = deployment_config["release_tag"]

    workflow_deployment_release = vellum_client.workflow_deployments.retrieve_workflow_deployment_release(
        deployment, release_tag
    )

    input_variables = workflow_deployment_release.workflow_version.input_variables
    description = workflow_deployment_release.description

    properties = {}
    required = []

    for input_var in input_variables:
        properties[input_var.key] = _compile_workflow_deployment_input(input_var)

        if input_var.required and input_var.default is None:
            required.append(input_var.key)

    parameters = {"type": "object", "properties": properties, "required": required}

    return FunctionDefinition(
        name=deployment.replace("-", ""),
        description=description,
        parameters=parameters,
    )



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .test_suite_test_case_delete_bulk_operation_data_request import TestSuiteTestCaseDeleteBulkOperationDataRequest


class TestSuiteTestCaseDeleteBulkOperationRequest(UniversalBaseModel):
    """
    A bulk operation that represents the deletion of a Test Case.
    """

    id: str = pydantic.Field()
    """
    An ID representing this specific operation. Can later be used to look up information about the operation's success in the response.
    """

    type: typing.Literal["DELETE"] = "DELETE"
    data: TestSuiteTestCaseDeleteBulkOperationDataRequest = pydantic.Field()
    """
    Information about the Test Case to delete
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from ...core.request_options import RequestOptions
from ...core.serialization import convert_and_respect_annotation_metadata
from ...types.paginated_test_suite_run_execution_list import PaginatedTestSuiteRunExecutionList
from ...types.test_suite_run_exec_config_request import TestSuiteRunExecConfigRequest
from ...types.test_suite_run_read import TestSuiteRunRead

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawTestSuiteRunsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create(
        self,
        *,
        exec_config: TestSuiteRunExecConfigRequest,
        test_suite_id: typing.Optional[str] = OMIT,
        test_suite_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TestSuiteRunRead]:
        """
        Trigger a Test Suite and create a new Test Suite Run

        Parameters
        ----------
        exec_config : TestSuiteRunExecConfigRequest
            Configuration that defines how the Test Suite should be run

        test_suite_id : typing.Optional[str]
            The ID of the Test Suite to run. Must provide either this or test_suite_id.

        test_suite_name : typing.Optional[str]
            The name of the Test Suite to run. Must provide either this or test_suite_id.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TestSuiteRunRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/test-suite-runs",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "test_suite_id": test_suite_id,
                "test_suite_name": test_suite_name,
                "exec_config": convert_and_respect_annotation_metadata(
                    object_=exec_config, annotation=TestSuiteRunExecConfigRequest, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteRunRead,
                    parse_obj_as(
                        type_=TestSuiteRunRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[TestSuiteRunRead]:
        """
        Retrieve a specific Test Suite Run by ID

        Parameters
        ----------
        id : str
            A UUID string identifying this test suite run.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TestSuiteRunRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/test-suite-runs/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteRunRead,
                    parse_obj_as(
                        type_=TestSuiteRunRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def list_executions(
        self,
        id: str,
        *,
        expand: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PaginatedTestSuiteRunExecutionList]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this test suite run.

        expand : typing.Optional[typing.Union[str, typing.Sequence[str]]]
            The response fields to expand for more information.
            - 'results.metric_results.metric_label' expands the metric label for each metric result.
            - 'results.metric_results.metric_definition' expands the metric definition for each metric result.
            - 'results.metric_results.metric_definition.name' expands the metric definition name for each metric result.

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PaginatedTestSuiteRunExecutionList]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/test-suite-runs/{jsonable_encoder(id)}/executions",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "expand": expand,
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedTestSuiteRunExecutionList,
                    parse_obj_as(
                        type_=PaginatedTestSuiteRunExecutionList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawTestSuiteRunsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create(
        self,
        *,
        exec_config: TestSuiteRunExecConfigRequest,
        test_suite_id: typing.Optional[str] = OMIT,
        test_suite_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TestSuiteRunRead]:
        """
        Trigger a Test Suite and create a new Test Suite Run

        Parameters
        ----------
        exec_config : TestSuiteRunExecConfigRequest
            Configuration that defines how the Test Suite should be run

        test_suite_id : typing.Optional[str]
            The ID of the Test Suite to run. Must provide either this or test_suite_id.

        test_suite_name : typing.Optional[str]
            The name of the Test Suite to run. Must provide either this or test_suite_id.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TestSuiteRunRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/test-suite-runs",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "test_suite_id": test_suite_id,
                "test_suite_name": test_suite_name,
                "exec_config": convert_and_respect_annotation_metadata(
                    object_=exec_config, annotation=TestSuiteRunExecConfigRequest, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteRunRead,
                    parse_obj_as(
                        type_=TestSuiteRunRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[TestSuiteRunRead]:
        """
        Retrieve a specific Test Suite Run by ID

        Parameters
        ----------
        id : str
            A UUID string identifying this test suite run.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TestSuiteRunRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/test-suite-runs/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteRunRead,
                    parse_obj_as(
                        type_=TestSuiteRunRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def list_executions(
        self,
        id: str,
        *,
        expand: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PaginatedTestSuiteRunExecutionList]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this test suite run.

        expand : typing.Optional[typing.Union[str, typing.Sequence[str]]]
            The response fields to expand for more information.
            - 'results.metric_results.metric_label' expands the metric label for each metric result.
            - 'results.metric_results.metric_definition' expands the metric definition for each metric result.
            - 'results.metric_results.metric_definition.name' expands the metric definition name for each metric result.

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PaginatedTestSuiteRunExecutionList]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/test-suite-runs/{jsonable_encoder(id)}/executions",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "expand": expand,
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedTestSuiteRunExecutionList,
                    parse_obj_as(
                        type_=PaginatedTestSuiteRunExecutionList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_search_node.workflow import BasicSearchWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow with a search node
    # WHEN we serialize it

    workflow_display = get_workflow_display(workflow_class=BasicSearchWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert input_variables == [
        {
            "id": "6e405c6c-36eb-4c06-9d54-ae06cccce585",
            "key": "query",
            "type": "STRING",
            "default": None,
            "required": True,
            "extensions": {"color": None},
        }
    ]

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [{"id": "27424f7d-9767-4059-bdcf-c2be8b798fd7", "key": "text", "type": "STRING"}]

    # AND its raw data is what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["nodes"]) == 3
    assert len(workflow_raw_data["edges"]) == 2

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "06671b25-5c6b-4675-8c74-6c396a608728",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "df80b4aa-2ba1-49a2-8375-fb1f78eee31f"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    search_node = workflow_raw_data["nodes"][1]
    assert search_node == {
        "id": "ab3a1413-c7b5-4cb0-a2d4-f5ab7d1d65b4",
        "type": "SEARCH",
        "inputs": [
            {
                "id": "49d21956-6e62-472b-b62a-7ec65faea1fd",
                "key": "query",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "6e405c6c-36eb-4c06-9d54-ae06cccce585"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "8cb28a69-088d-410d-bd0d-886d57ce7b64",
                "key": "document_index_id",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "name"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "983f2b7f-ad86-45cf-b04b-08724af27236",
                "key": "weights",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {"type": "JSON", "value": {"keywords": 0.45, "semantic_similarity": 0.55}},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "8072ec05-5fe4-47db-bc48-4c20ce49e123",
                "key": "limit",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": None}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "051c5d2e-4667-4ae2-9202-1076b21adf7b",
                "key": "separator",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "\n\n#####\n\n"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "8aac5dac-209e-48f3-97e1-0a39e4cd98d5",
                "key": "result_merging_enabled",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "False"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "036dee8f-194a-4b92-9739-69c98a4aa1b9",
                "key": "external_id_filters",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": None}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "855d3f57-e633-467e-a348-a394360247df",
                "key": "metadata_filters",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {
                                "type": "JSON",
                                "value": {
                                    "type": "LOGICAL_CONDITION_GROUP",
                                    "combinator": "AND",
                                    "conditions": [
                                        {
                                            "type": "LOGICAL_CONDITION",
                                            "lhs_variable_id": "a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                                            "operator": "=",
                                            "rhs_variable_id": "c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                                        },
                                        {
                                            "type": "LOGICAL_CONDITION",
                                            "lhs_variable_id": "a89483b6-6850-4105-8c4e-ec0fd197cd43",
                                            "operator": "=",
                                            "rhs_variable_id": "847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                                        },
                                    ],
                                    "negated": False,
                                },
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                "key": "vellum-query-builder-variable-a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "b118247f-96dd-4b3e-8289-9f277483c520"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                "key": "vellum-query-builder-variable-c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "aae2c10a-88b7-40bd-87a2-5e1e60c1e906"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "a89483b6-6850-4105-8c4e-ec0fd197cd43",
                "key": "vellum-query-builder-variable-a89483b6-6850-4105-8c4e-ec0fd197cd43",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "c9611a62-d1f5-4b41-bf9c-1aa3355760b4"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                "key": "vellum-query-builder-variable-847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "f374640e-a5c0-470e-ac71-c36c2b198c00"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Simple Search Node",
            "results_output_id": "e27fa934-589a-48f7-92a9-dcc90710ec7b",
            "text_output_id": "3f3bd066-ce73-46ee-84f1-d8ece69ecd8c",
            "error_output_id": None,
            "source_handle_id": "00ae06b3-f8d9-4ae6-9fbf-e4ff4d520e9b",
            "target_handle_id": "6d50305f-588b-469f-a042-b0767d3f99b1",
            "query_node_input_id": "49d21956-6e62-472b-b62a-7ec65faea1fd",
            "document_index_node_input_id": "8cb28a69-088d-410d-bd0d-886d57ce7b64",
            "weights_node_input_id": "983f2b7f-ad86-45cf-b04b-08724af27236",
            "limit_node_input_id": "8072ec05-5fe4-47db-bc48-4c20ce49e123",
            "separator_node_input_id": "051c5d2e-4667-4ae2-9202-1076b21adf7b",
            "result_merging_enabled_node_input_id": "8aac5dac-209e-48f3-97e1-0a39e4cd98d5",
            "external_id_filters_node_input_id": "036dee8f-194a-4b92-9739-69c98a4aa1b9",
            "metadata_filters_node_input_id": "855d3f57-e633-467e-a348-a394360247df",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "SearchNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "search_node", "node"],
        },
        "definition": {
            "name": "SimpleSearchNode",
            "module": ["tests", "workflows", "basic_search_node", "workflow"],
        },
        "trigger": {
            "id": "6d50305f-588b-469f-a042-b0767d3f99b1",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "00ae06b3-f8d9-4ae6-9fbf-e4ff4d520e9b", "name": "default", "type": "DEFAULT"}],
    }

    final_output_node = workflow_raw_data["nodes"][2]
    assert final_output_node == {
        "id": "4e466510-6756-403f-a182-56e5a2b85d94",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "text",
            "target_handle_id": "cd8c736f-1b77-493d-b857-d8feb5c03b15",
            "output_id": "27424f7d-9767-4059-bdcf-c2be8b798fd7",
            "output_type": "STRING",
            "node_input_id": "39c7f674-a794-4525-8a04-f22a40ed0914",
        },
        "inputs": [
            {
                "id": "39c7f674-a794-4525-8a04-f22a40ed0914",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "ab3a1413-c7b5-4cb0-a2d4-f5ab7d1d65b4",
                                "output_id": "3f3bd066-ce73-46ee-84f1-d8ece69ecd8c",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 400.0, "y": -50.0}},
        "base": {
            "name": "FinalOutputNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
        },
        "definition": None,
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "06533904-4897-4a7a-aa8d-50419b3d33ae",
            "source_node_id": "06671b25-5c6b-4675-8c74-6c396a608728",
            "source_handle_id": "df80b4aa-2ba1-49a2-8375-fb1f78eee31f",
            "target_node_id": "ab3a1413-c7b5-4cb0-a2d4-f5ab7d1d65b4",
            "target_handle_id": "6d50305f-588b-469f-a042-b0767d3f99b1",
            "type": "DEFAULT",
        },
        {
            "id": "cb918deb-f546-47b5-8b6b-db0d22a29fd1",
            "source_node_id": "ab3a1413-c7b5-4cb0-a2d4-f5ab7d1d65b4",
            "source_handle_id": "00ae06b3-f8d9-4ae6-9fbf-e4ff4d520e9b",
            "target_node_id": "4e466510-6756-403f-a182-56e5a2b85d94",
            "target_handle_id": "cd8c736f-1b77-493d-b857-d8feb5c03b15",
            "type": "DEFAULT",
        },
    ]

    # AND the display data is what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}}

    # AND the definition is what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicSearchWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_search_node",
            "workflow",
        ],
    }



from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, Any, Dict, List, cast

from pydantic import BaseModel

from vellum.client.types.logical_operator import LogicalOperator
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.expressions.add import AddExpression
from vellum.workflows.expressions.and_ import AndExpression
from vellum.workflows.expressions.begins_with import BeginsWithExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.expressions.contains import ContainsExpression
from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression
from vellum.workflows.expressions.ends_with import EndsWithExpression
from vellum.workflows.expressions.equals import EqualsExpression
from vellum.workflows.expressions.greater_than import GreaterThanExpression
from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
from vellum.workflows.expressions.in_ import InExpression
from vellum.workflows.expressions.is_nil import IsNilExpression
from vellum.workflows.expressions.is_not_nil import IsNotNilExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.is_undefined import IsUndefinedExpression
from vellum.workflows.expressions.length import LengthExpression
from vellum.workflows.expressions.less_than import LessThanExpression
from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
from vellum.workflows.expressions.minus import MinusExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.not_in import NotInExpression
from vellum.workflows.expressions.or_ import OrExpression
from vellum.workflows.expressions.parse_json import ParseJsonExpression
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.references.environment_variable import EnvironmentVariableReference
from vellum.workflows.references.execution_count import ExecutionCountReference
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.references.output import OutputReference
from vellum.workflows.references.state_value import StateValueReference
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum.workflows.references.workflow_input import WorkflowInputReference
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum.workflows.types.definition import DeploymentDefinition
from vellum.workflows.types.generics import is_workflow_class
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.utils.exceptions import UnsupportedSerializationException

if TYPE_CHECKING:
    from vellum_ee.workflows.display.types import WorkflowDisplayContext


def convert_descriptor_to_operator(descriptor: BaseDescriptor) -> LogicalOperator:
    if isinstance(descriptor, EqualsExpression):
        return "="
    elif isinstance(descriptor, DoesNotEqualExpression):
        return "!="
    elif isinstance(descriptor, LessThanExpression):
        return "<"
    elif isinstance(descriptor, GreaterThanExpression):
        return ">"
    elif isinstance(descriptor, LessThanOrEqualToExpression):
        return "<="
    elif isinstance(descriptor, GreaterThanOrEqualToExpression):
        return ">="
    elif isinstance(descriptor, ContainsExpression):
        return "contains"
    elif isinstance(descriptor, BeginsWithExpression):
        return "beginsWith"
    elif isinstance(descriptor, EndsWithExpression):
        return "endsWith"
    elif isinstance(descriptor, DoesNotContainExpression):
        return "doesNotContain"
    elif isinstance(descriptor, DoesNotBeginWithExpression):
        return "doesNotBeginWith"
    elif isinstance(descriptor, DoesNotEndWithExpression):
        return "doesNotEndWith"
    elif isinstance(descriptor, (IsNullExpression, IsNilExpression, IsUndefinedExpression)):
        return "null"
    elif isinstance(descriptor, (IsNotNullExpression, IsNotNilExpression, IsNotUndefinedExpression)):
        return "notNull"
    elif isinstance(descriptor, InExpression):
        return "in"
    elif isinstance(descriptor, NotInExpression):
        return "notIn"
    elif isinstance(descriptor, BetweenExpression):
        return "between"
    elif isinstance(descriptor, NotBetweenExpression):
        return "notBetween"
    elif isinstance(descriptor, AndExpression):
        return "and"
    elif isinstance(descriptor, OrExpression):
        return "or"
    elif isinstance(descriptor, CoalesceExpression):
        return "coalesce"
    elif isinstance(descriptor, ParseJsonExpression):
        return "parseJson"
    elif isinstance(descriptor, LengthExpression):
        return "length"
    elif isinstance(descriptor, AddExpression):
        return "+"
    elif isinstance(descriptor, MinusExpression):
        return "-"
    elif isinstance(descriptor, ConcatExpression):
        return "concat"
    else:
        raise ValueError(f"Unsupported descriptor type: {descriptor}")


def get_child_descriptor(value: LazyReference, display_context: "WorkflowDisplayContext") -> BaseDescriptor:
    if isinstance(value._get, str):
        reference_parts = value._get.split(".")
        if len(reference_parts) < 3:
            raise Exception(f"Failed to parse lazy reference: {value._get}. Only Node Output references are supported.")

        output_name = reference_parts[-1]
        nested_class_name = reference_parts[-2]
        if nested_class_name != "Outputs":
            raise Exception(
                f"Failed to parse lazy reference: {value._get}. Outputs are the only node reference supported."
            )

        node_class_name = ".".join(reference_parts[:-2])
        for node in display_context.global_node_displays.keys():
            if node.__name__ == node_class_name:
                return getattr(node.Outputs, output_name)

        raise Exception(f"Failed to parse lazy reference: {value._get}")

    return value._get()


def _serialize_condition(display_context: "WorkflowDisplayContext", condition: BaseDescriptor) -> JsonObject:
    if isinstance(
        condition,
        (
            IsNullExpression,
            IsNotNullExpression,
            IsNilExpression,
            IsNotNilExpression,
            IsUndefinedExpression,
            IsNotUndefinedExpression,
            LengthExpression,
            ParseJsonExpression,
        ),
    ):
        lhs = serialize_value(display_context, condition._expression)
        return {
            "type": "UNARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
        }
    elif isinstance(condition, (BetweenExpression, NotBetweenExpression)):
        base = serialize_value(display_context, condition._value)
        lhs = serialize_value(display_context, condition._start)
        rhs = serialize_value(display_context, condition._end)

        return {
            "type": "TERNARY_EXPRESSION",
            "base": base,
            "operator": convert_descriptor_to_operator(condition),
            "lhs": lhs,
            "rhs": rhs,
        }
    elif isinstance(
        condition,
        (
            AddExpression,
            AndExpression,
            BeginsWithExpression,
            CoalesceExpression,
            ConcatExpression,
            ContainsExpression,
            DoesNotBeginWithExpression,
            DoesNotContainExpression,
            DoesNotEndWithExpression,
            DoesNotEqualExpression,
            EndsWithExpression,
            EqualsExpression,
            GreaterThanExpression,
            GreaterThanOrEqualToExpression,
            InExpression,
            LessThanExpression,
            LessThanOrEqualToExpression,
            MinusExpression,
            NotInExpression,
            OrExpression,
        ),
    ):
        lhs = serialize_value(display_context, condition._lhs)
        rhs = serialize_value(display_context, condition._rhs)

        return {
            "type": "BINARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
            "rhs": rhs,
        }
    elif isinstance(condition, AccessorExpression):
        return {
            "type": "BINARY_EXPRESSION",
            "lhs": serialize_value(display_context, condition._base),
            "operator": "accessField",
            "rhs": serialize_value(display_context, condition._field),
        }

    raise UnsupportedSerializationException(f"Unsupported condition type: {condition.__class__.__name__}")


def serialize_key(key: Any) -> str:
    """Serialize dictionary keys to strings, handling function objects properly."""
    if callable(key):
        return key.__name__
    elif isinstance(key, str):
        return key
    else:
        return str(key)


def serialize_value(display_context: "WorkflowDisplayContext", value: Any) -> JsonObject:
    if isinstance(value, ConstantValueReference):
        return serialize_value(display_context, value._value)

    if isinstance(value, LazyReference):
        child_descriptor = get_child_descriptor(value, display_context)
        return serialize_value(display_context, child_descriptor)

    if isinstance(value, WorkflowInputReference):
        try:
            workflow_input_display = display_context.global_workflow_input_displays[value]
        except KeyError:
            inputs_class_name = value.inputs_class.__name__
            workflow_class_name = display_context.workflow_display_class.infer_workflow_class().__name__
            raise UnsupportedSerializationException(
                f"Inputs class '{inputs_class_name}' referenced during serialization of '{workflow_class_name}' "
                f"without parameterizing this Workflow with this Inputs definition. Update your Workflow "
                f"definition to '{workflow_class_name}(BaseWorkflow[{inputs_class_name}, BaseState])'."
            )
        return {
            "type": "WORKFLOW_INPUT",
            "input_variable_id": str(workflow_input_display.id),
        }

    if isinstance(value, StateValueReference):
        state_value_display = display_context.global_state_value_displays[value]
        return {
            "type": "WORKFLOW_STATE",
            "state_variable_id": str(state_value_display.id),
        }

    if isinstance(value, OutputReference):
        upstream_node, output_display = display_context.global_node_output_displays[value]
        upstream_node_display = display_context.global_node_displays[upstream_node]

        return {
            "type": "NODE_OUTPUT",
            "node_id": str(upstream_node_display.node_id),
            "node_output_id": str(output_display.id),
        }

    if isinstance(value, VellumSecretReference):
        return {
            "type": "VELLUM_SECRET",
            "vellum_secret_name": value.name,
        }

    if isinstance(value, EnvironmentVariableReference):
        if value.serialize_as_constant:
            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "STRING",
                    "value": value.name,
                },
            }
        else:
            return {
                "type": "ENVIRONMENT_VARIABLE",
                "environment_variable": value.name,
            }

    if isinstance(value, ExecutionCountReference):
        node_class_display = display_context.global_node_displays[value.node_class]

        return {
            "type": "EXECUTION_COUNTER",
            "node_id": str(node_class_display.node_id),
        }

    if isinstance(value, list):
        serialized_items = [serialize_value(display_context, item) for item in value]
        if all(isinstance(item, dict) and item["type"] == "CONSTANT_VALUE" for item in serialized_items):
            constant_values = []
            for item in serialized_items:
                item_dict = cast(Dict[str, Any], item)
                value_inner = item_dict["value"]

                if value_inner["type"] == "JSON" and "items" in value_inner:
                    # Nested JSON list
                    constant_values.append(value_inner["items"])
                else:
                    constant_values.append(value_inner["value"])

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_values,
                },
            }
        else:
            return {
                "type": "ARRAY_REFERENCE",
                "items": cast(JsonArray, serialized_items),  # list[JsonObject] -> JsonArray
            }

    if is_dataclass(value) and not isinstance(value, type):
        dict_value = asdict(value)
        return serialize_value(display_context, dict_value)

    if isinstance(value, dict):
        serialized_entries: List[Dict[str, Any]] = [
            {
                "id": str(uuid4_from_hash(f"{key}|{val}")),
                "key": serialize_key(key),
                "value": serialize_value(display_context, val),
            }
            for key, val in value.items()
        ]

        # Check if all entries have constant values
        if all(entry["value"]["type"] == "CONSTANT_VALUE" for entry in serialized_entries):
            constant_entries = {}
            for entry in serialized_entries:
                entry_value = entry["value"]["value"]
                constant_entries[entry["key"]] = entry_value["value"]

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_entries,
                },
            }
        else:
            return {"type": "DICTIONARY_REFERENCE", "entries": cast(JsonArray, serialized_entries)}

    if is_workflow_class(value):
        from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

        # Pass the parent display context so the subworkflow can resolve parent workflow inputs
        workflow_display = get_workflow_display(
            workflow_class=value,
            parent_display_context=display_context,
        )
        serialized_value: dict = workflow_display.serialize()
        name = serialized_value["workflow_raw_data"]["definition"]["name"]
        description = value.__doc__ or ""
        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": {
                    "type": "INLINE_WORKFLOW",
                    "name": name,
                    "description": description,
                    "exec_config": serialized_value,
                },
            },
        }

    if isinstance(value, DeploymentDefinition):
        workflow_deployment_release = display_context.client.workflow_deployments.retrieve_workflow_deployment_release(
            value.deployment, value.release_tag
        )
        name = workflow_deployment_release.deployment.name or value.deployment
        description = workflow_deployment_release.description or f"Workflow Deployment for {name}"

        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": {
                    "type": "WORKFLOW_DEPLOYMENT",
                    "name": name,
                    "description": description,
                    "deployment": value.deployment,
                    "release_tag": value.release_tag,
                },
            },
        }

    if isinstance(value, BaseModel):
        dict_value = value.model_dump()
        return serialize_value(display_context, dict_value)

    if not isinstance(value, BaseDescriptor):
        vellum_value = primitive_to_vellum_value(value)
        return {
            "type": "CONSTANT_VALUE",
            "value": vellum_value.dict(),
        }

    # If it's not any of the references we know about,
    # then try to serialize it as a nested value
    return _serialize_condition(display_context, value)



from functools import lru_cache
from typing import Any, Dict, Literal, NamedTuple, Optional, Tuple, Union
from typing_extensions import TypeAlias

from pydantic.fields import FieldInfo
from pydantic.plugin import (
    PydanticPluginProtocol,
    ValidateJsonHandlerProtocol,
    ValidatePythonHandlerProtocol,
    ValidateStringsHandlerProtocol,
)
from pydantic_core import CoreSchema

# Redefined manually instead of imported from pydantic to support versions < 2.5
SchemaKind: TypeAlias = Literal["BaseModel", "TypeAdapter", "dataclass", "create_model", "validate_call"]


# Redefined manually instead of imported from pydantic to support versions < 2.5
class SchemaTypePath(NamedTuple):
    """Path defining where `schema_type` was defined, or where `TypeAdapter` was called."""

    module: str
    name: str


@lru_cache(maxsize=1)
def import_base_descriptor():
    """
    We have to avoid importing from vellum.* in this file because it will cause a circular import.
    """
    try:
        from vellum.workflows.descriptors.base import BaseDescriptor

        return BaseDescriptor
    except Exception:
        return None


# https://docs.pydantic.dev/2.8/concepts/plugins/#build-a-plugin
class OnValidatePython(ValidatePythonHandlerProtocol):
    tracked_descriptors: Dict[str, Any] = {}

    def on_enter(
        self,
        input: Any,
        *,
        strict: Optional[bool] = None,
        from_attributes: Optional[bool] = None,
        context: Optional[Dict[str, Any]] = None,
        self_instance: Optional[Any] = None,
        allow_partial: Union[bool, Literal["off", "on", "trailing-strings"]] = False,
        by_alias: Optional[bool] = None,
        by_name: Optional[bool] = None,
    ) -> None:
        if not isinstance(input, dict):
            return

        if self_instance:
            model_fields: Dict[str, FieldInfo] = self_instance.__class__.model_fields
        else:
            model_fields = {}

        self.tracked_descriptors = {}
        BaseDescriptor = import_base_descriptor()

        # If BaseDescriptor import failed, skip descriptor processing
        if BaseDescriptor is None:
            return

        for key, value in input.items():
            field_info = model_fields.get(key)
            if isinstance(value, BaseDescriptor) and (
                not field_info
                or not field_info.annotation
                or not isinstance(field_info.annotation, type)
                or not issubclass(field_info.annotation, BaseDescriptor)
            ):
                self.tracked_descriptors[key] = value
                # TODO: This does not yet work for descriptors that map to more complex types
                # https://app.shortcut.com/vellum/story/4636
                if len(value.types) == 0:
                    input[key] = dict()
                else:
                    input[key] = value.types[0]()

    def on_success(self, result: Any) -> None:
        if self.tracked_descriptors:
            frozen = result.model_config.get("frozen")
            if frozen:
                result.model_config["frozen"] = False

            for key, value in self.tracked_descriptors.items():
                setattr(result, key, value)

            if frozen:
                result.model_config["frozen"] = True

            self.tracked_descriptors = {}


class VellumPydanticPlugin(PydanticPluginProtocol):
    def new_schema_validator(
        self,
        schema: CoreSchema,
        schema_type: Any,
        schema_type_path: SchemaTypePath,  # type: ignore
        schema_kind: SchemaKind,
        config: Any,
        plugin_settings: Dict[str, Any],
    ) -> Tuple[
        Union[ValidatePythonHandlerProtocol, None],
        Union[ValidateJsonHandlerProtocol, None],
        Union[ValidateStringsHandlerProtocol, None],
    ]:
        return OnValidatePython(), None, None


pydantic_plugin = VellumPydanticPlugin()



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import DeploymentsListRequestStatus, ListDeploymentReleaseTagsRequestSource

__all__ = ["DeploymentsListRequestStatus", "ListDeploymentReleaseTagsRequestSource"]



from typing import Any

from pydantic import BaseModel

from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.state.base import BaseState


class TestState(BaseState):
    pass


class TestModel(BaseModel):
    field: Any


def test_pydantic_plugin__empty_types_descriptor():
    """
    Test that the pydantic plugin handles BaseDescriptor with empty types gracefully.
    """

    base_ref = ConstantValueReference({"name": "test"})
    accessor = AccessorExpression(base=base_ref, field="name")

    model = TestModel(field=accessor)

    assert model is not None
    assert hasattr(model, "field")
    assert model.field == accessor



# This file was auto-generated by Fern from our API Definition.

import typing

WorkflowNodeResultEventState = typing.Union[
    typing.Literal["INITIATED", "STREAMING", "FULFILLED", "REJECTED"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class WorkflowExpandMetaRequest(UniversalBaseModel):
    usage: typing.Optional[bool] = pydantic.Field(default=None)
    """
    If enabled, the Prompt Node FULFILLED events will include model host usage tracking. This may increase latency for some model hosts.
    """

    cost: typing.Optional[bool] = pydantic.Field(default=None)
    """
    If enabled, the Prompt Node FULFILLED events will include model host cost tracking. This may increase latency for some model hosts.
    """

    model_name: typing.Optional[bool] = pydantic.Field(default=None)
    """
    If enabled, the Prompt Node FULFILLED events will include model host name
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import pytest
from uuid import UUID

from vellum.workflows.utils.uuids import uuid4_from_hash


@pytest.mark.parametrize(
    ["input_str", "expected"],
    [
        ("MyExampleString", UUID("b2dadec1-bff4-4e26-8d65-e99e62628cd2")),
        ("My Example String", UUID("a1e68bde-3263-4526-88bd-70f4bf800224")),
    ],
)
def test_uuid4_from_hash(input_str, expected):
    actual = uuid4_from_hash(input_str)
    assert actual == expected



import pytest

from vellum.workflows.constants import undefined


def test_undefined__ensure_sensible_error_messages():
    # WHEN we invoke an invalid operation on `undefined`
    with pytest.raises(Exception) as e:
        len(undefined)

    # THEN we get a sensible error message
    assert str(e.value) == "object of type 'undefined' has no len()"



from vellum.workflows.nodes.displayable import FinalOutputNode
from vellum.workflows.state import BaseState

from .templating_node import TemplatingNode


class FinalOutput(FinalOutputNode[BaseState, str]):
    class Outputs(FinalOutputNode.Outputs):
        value = TemplatingNode.Outputs.result



import os
from typing import TYPE_CHECKING, Any, Optional

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor

if TYPE_CHECKING:
    from vellum.workflows.state.base import BaseState


class EnvironmentVariableReference(BaseDescriptor[str]):
    def __init__(
        self,
        *,
        name: str,
        # DEPRECATED - to be removed in 0.15.0 release
        default: Optional[str] = None,
        serialize_as_constant: bool = False,
    ):
        super().__init__(name=name, types=(str,))
        self._serialize_as_constant = serialize_as_constant

    def resolve(self, state: "BaseState") -> Any:
        env_value = os.environ.get(self.name)
        if env_value is not None:
            return env_value

        return undefined

    @property
    def serialize_as_constant(self) -> bool:
        return self._serialize_as_constant

    @serialize_as_constant.setter
    def serialize_as_constant(self, value: bool):
        self._serialize_as_constant = value



from abc import ABC, abstractmethod
from typing import Iterator

from vellum.workflows.events.workflow import WorkflowEvent
from vellum.workflows.state.base import BaseState


class BaseWorkflowResolver(ABC):
    @abstractmethod
    def get_latest_execution_events(self) -> Iterator[WorkflowEvent]:
        pass

    @abstractmethod
    def get_state_snapshot_history(self) -> Iterator[BaseState]:
        pass



from typing import Any, ClassVar, Dict, Generic, TypeVar, Union

from vellum.workflows.nodes import NoteNode
from vellum.workflows.types.core import JsonObject
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_NoteNodeType = TypeVar("_NoteNodeType", bound=NoteNode)


class BaseNoteNodeDisplay(BaseNodeDisplay[_NoteNodeType], Generic[_NoteNodeType]):
    text: ClassVar[str] = ""
    style: ClassVar[Union[Dict[str, Any], None]] = None

    def serialize(self, display_context: WorkflowDisplayContext, **kwargs: Any) -> JsonObject:
        del kwargs  # Unused parameters
        node_id = self.node_id

        return {
            "id": str(node_id),
            "type": "NOTE",
            "inputs": [],
            "data": {
                "label": self.label,
                "text": self.text,
                "style": self.style,
            },
            **self.serialize_generic_fields(display_context),
        }



from typing import TypeVar, Union

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.utils import resolve_value
from vellum.workflows.state.base import BaseState
from vellum.workflows.types.utils import resolve_combined_types

LHS = TypeVar("LHS")
RHS = TypeVar("RHS")


class CoalesceExpression(BaseDescriptor[Union[LHS, RHS]]):
    def __init__(
        self,
        *,
        lhs: Union[BaseDescriptor[LHS], LHS],
        rhs: Union[BaseDescriptor[RHS], RHS],
    ) -> None:
        super().__init__(
            name=f"{lhs} ?? {rhs}",
            types=resolve_combined_types(lhs, rhs),
            instance=None,
        )
        self._lhs = lhs
        self._rhs = rhs

    def resolve(self, state: "BaseState") -> Union[LHS, RHS]:
        lhs = resolve_value(self._lhs, state)
        if lhs is not undefined and lhs is not None:
            return lhs

        return resolve_value(self._rhs, state)

    @property
    def lhs(self) -> Union[BaseDescriptor[LHS], LHS]:
        return self._lhs

    @property
    def rhs(self) -> Union[BaseDescriptor[RHS], RHS]:
        return self._rhs



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.paginated_workflow_sandbox_example_list import PaginatedWorkflowSandboxExampleList
from ...types.workflow_deployment_read import WorkflowDeploymentRead
from .raw_client import AsyncRawWorkflowSandboxesClient, RawWorkflowSandboxesClient
from .types.list_workflow_sandbox_examples_request_tag import ListWorkflowSandboxExamplesRequestTag

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class WorkflowSandboxesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawWorkflowSandboxesClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawWorkflowSandboxesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawWorkflowSandboxesClient
        """
        return self._raw_client

    def deploy_workflow(
        self,
        id: str,
        workflow_id: str,
        *,
        workflow_deployment_id: typing.Optional[str] = OMIT,
        workflow_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowDeploymentRead:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this workflow sandbox.

        workflow_id : str
            An ID identifying the Workflow you'd like to deploy.

        workflow_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Workflow Deployment you'd like to update. Cannot specify both this and workflow_deployment_name. Leave null to create a new Workflow Deployment.

        workflow_deployment_name : typing.Optional[str]
            The unique name of the Workflow Deployment you'd like to either create or update. Cannot specify both this and workflow_deployment_id. If provided and matches an existing Workflow Deployment, that Workflow Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Workflow Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_sandboxes.deploy_workflow(
            id="id",
            workflow_id="workflow_id",
        )
        """
        _response = self._raw_client.deploy_workflow(
            id,
            workflow_id,
            workflow_deployment_id=workflow_deployment_id,
            workflow_deployment_name=workflow_deployment_name,
            label=label,
            release_tags=release_tags,
            release_description=release_description,
            request_options=request_options,
        )
        return _response.data

    def list_workflow_sandbox_examples(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        tag: typing.Optional[ListWorkflowSandboxExamplesRequestTag] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedWorkflowSandboxExampleList:
        """
        List Workflow Sandbox examples that were previously cloned into the User's Workspace

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        tag : typing.Optional[ListWorkflowSandboxExamplesRequestTag]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedWorkflowSandboxExampleList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_sandboxes.list_workflow_sandbox_examples()
        """
        _response = self._raw_client.list_workflow_sandbox_examples(
            limit=limit, offset=offset, ordering=ordering, tag=tag, request_options=request_options
        )
        return _response.data


class AsyncWorkflowSandboxesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawWorkflowSandboxesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawWorkflowSandboxesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawWorkflowSandboxesClient
        """
        return self._raw_client

    async def deploy_workflow(
        self,
        id: str,
        workflow_id: str,
        *,
        workflow_deployment_id: typing.Optional[str] = OMIT,
        workflow_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowDeploymentRead:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this workflow sandbox.

        workflow_id : str
            An ID identifying the Workflow you'd like to deploy.

        workflow_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Workflow Deployment you'd like to update. Cannot specify both this and workflow_deployment_name. Leave null to create a new Workflow Deployment.

        workflow_deployment_name : typing.Optional[str]
            The unique name of the Workflow Deployment you'd like to either create or update. Cannot specify both this and workflow_deployment_id. If provided and matches an existing Workflow Deployment, that Workflow Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Workflow Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_sandboxes.deploy_workflow(
                id="id",
                workflow_id="workflow_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.deploy_workflow(
            id,
            workflow_id,
            workflow_deployment_id=workflow_deployment_id,
            workflow_deployment_name=workflow_deployment_name,
            label=label,
            release_tags=release_tags,
            release_description=release_description,
            request_options=request_options,
        )
        return _response.data

    async def list_workflow_sandbox_examples(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        tag: typing.Optional[ListWorkflowSandboxExamplesRequestTag] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedWorkflowSandboxExampleList:
        """
        List Workflow Sandbox examples that were previously cloned into the User's Workspace

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        tag : typing.Optional[ListWorkflowSandboxExamplesRequestTag]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedWorkflowSandboxExampleList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_sandboxes.list_workflow_sandbox_examples()


        asyncio.run(main())
        """
        _response = await self._raw_client.list_workflow_sandbox_examples(
            limit=limit, offset=offset, ordering=ordering, tag=tag, request_options=request_options
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .execution_vellum_value import ExecutionVellumValue
from .workflow_error import WorkflowError
from .workflow_execution_actual import WorkflowExecutionActual
from .workflow_execution_usage_result import WorkflowExecutionUsageResult
from .workflow_execution_view_online_eval_metric_result import WorkflowExecutionViewOnlineEvalMetricResult


class SlimWorkflowExecutionRead(UniversalBaseModel):
    span_id: str
    parent_context: typing.Optional["WorkflowDeploymentParentContext"] = None
    start: dt.datetime
    end: typing.Optional[dt.datetime] = None
    inputs: typing.List[ExecutionVellumValue]
    outputs: typing.List[ExecutionVellumValue]
    error: typing.Optional[WorkflowError] = None
    latest_actual: typing.Optional[WorkflowExecutionActual] = None
    metric_results: typing.List[WorkflowExecutionViewOnlineEvalMetricResult]
    usage_results: typing.Optional[typing.List[WorkflowExecutionUsageResult]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .api_request_parent_context import ApiRequestParentContext  # noqa: E402, F401, I001
from .external_parent_context import ExternalParentContext  # noqa: E402, F401, I001
from .node_parent_context import NodeParentContext  # noqa: E402, F401, I001
from .prompt_deployment_parent_context import PromptDeploymentParentContext  # noqa: E402, F401, I001
from .span_link import SpanLink  # noqa: E402, F401, I001
from .workflow_deployment_parent_context import WorkflowDeploymentParentContext  # noqa: E402, F401, I001
from .workflow_parent_context import WorkflowParentContext  # noqa: E402, F401, I001
from .workflow_sandbox_parent_context import WorkflowSandboxParentContext  # noqa: E402, F401, I001
from .array_vellum_value import ArrayVellumValue  # noqa: E402, F401, I001

update_forward_refs(SlimWorkflowExecutionRead)



from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay


def test_serialize_module_with_dataset():
    """Test that serialize_module correctly serializes dataset from sandbox modules."""
    module_path = "tests.workflows.basic_inputs_and_outputs"

    result = BaseWorkflowDisplay.serialize_module(module_path)

    assert hasattr(result, "dataset")

    assert result.dataset is None


def test_serialize_module_with_actual_dataset():
    """Test that serialize_module correctly serializes dataset when sandbox has dataset attribute."""
    module_path = "tests.workflows.test_dataset_serialization"

    result = BaseWorkflowDisplay.serialize_module(module_path)

    assert hasattr(result, "dataset")

    assert result.dataset is not None
    assert isinstance(result.dataset, list)
    assert len(result.dataset) == 2

    for i, item in enumerate(result.dataset):
        assert "label" in item
        assert "inputs" in item
        assert item["label"] == f"Scenario {i + 1}"
        assert isinstance(item["inputs"], dict)
        assert "message" in item["inputs"]


def test_serialize_module_happy_path():
    """Test that serialize_module works with a valid module path."""
    module_path = "tests.workflows.trivial"

    result = BaseWorkflowDisplay.serialize_module(module_path)

    assert hasattr(result, "exec_config")
    assert hasattr(result, "errors")
    assert isinstance(result.exec_config, dict)
    assert isinstance(result.errors, list)
    assert "workflow_raw_data" in result.exec_config
    assert "input_variables" in result.exec_config
    assert "output_variables" in result.exec_config


def test_serialize_module_includes_additional_files():
    """Test that serialize_module includes additional files from the module directory."""
    module_path = "tests.workflows.module_with_additional_files"

    result = BaseWorkflowDisplay.serialize_module(module_path)

    assert hasattr(result, "exec_config")
    assert hasattr(result, "errors")
    assert isinstance(result.exec_config, dict)
    assert isinstance(result.errors, list)

    assert "module_data" in result.exec_config
    assert "additional_files" in result.exec_config["module_data"]

    additional_files = result.exec_config["module_data"]["additional_files"]
    assert isinstance(additional_files, dict)

    assert "helper.py" in additional_files
    assert "data.txt" in additional_files
    assert "utils/constants.py" in additional_files

    assert "workflow.py" not in additional_files
    assert "__init__.py" not in additional_files
    assert "utils/__init__.py" in additional_files
    assert "nodes/test_node.py" not in additional_files

    assert "def helper_function():" in additional_files["helper.py"]
    assert "sample data file" in additional_files["data.txt"]
    assert "CONSTANT_VALUE" in additional_files["utils/constants.py"]



from contextlib import contextmanager
from dataclasses import field
import threading
from uuid import UUID
from typing import Iterator, Optional, cast

from vellum.client.core import UniversalBaseModel
from vellum.workflows.events.types import ParentContext


class ExecutionContext(UniversalBaseModel):
    trace_id: UUID = field(default_factory=lambda: UUID("00000000-0000-0000-0000-000000000000"))
    parent_context: Optional[ParentContext] = None


_CONTEXT_KEY = "_execution_context"

local = threading.local()


def get_execution_context() -> ExecutionContext:
    """Retrieve the current execution context."""
    return getattr(local, _CONTEXT_KEY, ExecutionContext())


def set_execution_context(context: ExecutionContext) -> None:
    """Set the current execution context."""
    setattr(local, _CONTEXT_KEY, context)


def get_parent_context() -> ParentContext:
    return cast(ParentContext, get_execution_context().parent_context)


@contextmanager
def execution_context(
    parent_context: Optional[ParentContext] = None, trace_id: Optional[UUID] = None
) -> Iterator[None]:
    """Context manager for handling execution context."""
    prev_context = get_execution_context()
    set_trace_id = (
        prev_context.trace_id
        if int(prev_context.trace_id)
        else trace_id or UUID("00000000-0000-0000-0000-000000000000")
    )
    set_parent_context = parent_context or prev_context.parent_context
    set_context = ExecutionContext(parent_context=set_parent_context, trace_id=set_trace_id)
    try:
        set_execution_context(set_context)
        yield
    finally:
        set_execution_context(prev_context)



from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
import logging
from queue import Empty, Queue
import sys
from threading import Event as ThreadingEvent, Thread
import traceback
from uuid import UUID, uuid4
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from vellum.workflows.constants import undefined
from vellum.workflows.context import ExecutionContext, execution_context, get_execution_context
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors import WorkflowError, WorkflowErrorCode
from vellum.workflows.events import (
    NodeExecutionFulfilledEvent,
    NodeExecutionInitiatedEvent,
    NodeExecutionRejectedEvent,
    NodeExecutionStreamingEvent,
    WorkflowEvent,
    WorkflowEventGenerator,
    WorkflowExecutionFulfilledEvent,
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionRejectedEvent,
    WorkflowExecutionStreamingEvent,
)
from vellum.workflows.events.node import (
    NodeExecutionFulfilledBody,
    NodeExecutionInitiatedBody,
    NodeExecutionRejectedBody,
    NodeExecutionStreamingBody,
)
from vellum.workflows.events.types import BaseEvent, NodeParentContext, ParentContext, WorkflowParentContext
from vellum.workflows.events.workflow import (
    WorkflowEventStream,
    WorkflowExecutionFulfilledBody,
    WorkflowExecutionInitiatedBody,
    WorkflowExecutionPausedBody,
    WorkflowExecutionPausedEvent,
    WorkflowExecutionRejectedBody,
    WorkflowExecutionResumedBody,
    WorkflowExecutionResumedEvent,
    WorkflowExecutionSnapshottedBody,
    WorkflowExecutionSnapshottedEvent,
    WorkflowExecutionStreamingBody,
)
from vellum.workflows.exceptions import NodeException, WorkflowInitializationException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.bases.base import NodeRunResponse
from vellum.workflows.nodes.mocks import MockNodeExecutionArg
from vellum.workflows.outputs import BaseOutputs
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.ports.port import Port
from vellum.workflows.references import ExternalInputReference, OutputReference
from vellum.workflows.references.state_value import StateValueReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.state.delta import StateDelta
from vellum.workflows.types.generics import InputsType, OutputsType, StateType

if TYPE_CHECKING:
    from vellum.workflows import BaseWorkflow

logger = logging.getLogger(__name__)

RunFromNodeArg = Sequence[Type[BaseNode]]
ExternalInputsArg = Dict[ExternalInputReference, Any]
BackgroundThreadItem = Union[BaseState, WorkflowEvent, None]


@dataclass
class ActiveNode(Generic[StateType]):
    node: BaseNode[StateType]
    was_outputs_streamed: bool = False


class WorkflowRunner(Generic[StateType]):
    _entrypoints: Iterable[Type[BaseNode]]

    def __init__(
        self,
        workflow: "BaseWorkflow[InputsType, StateType]",
        inputs: Optional[InputsType] = None,
        state: Optional[StateType] = None,
        entrypoint_nodes: Optional[RunFromNodeArg] = None,
        external_inputs: Optional[ExternalInputsArg] = None,
        cancel_signal: Optional[ThreadingEvent] = None,
        node_output_mocks: Optional[MockNodeExecutionArg] = None,
        max_concurrency: Optional[int] = None,
        init_execution_context: Optional[ExecutionContext] = None,
    ):
        if state and external_inputs:
            raise ValueError("Can only run a Workflow providing one of state or external inputs, not both")

        self.workflow = workflow
        self._is_resuming = False
        self._should_emit_initial_state = True
        if entrypoint_nodes:
            if len(list(entrypoint_nodes)) > 1:
                raise ValueError("Cannot resume from multiple nodes")

            # TODO: Support resuming from multiple nodes
            # https://app.shortcut.com/vellum/story/4408
            node = next(iter(entrypoint_nodes))
            if state:
                self._initial_state = deepcopy(state)
                self._initial_state.meta.span_id = uuid4()
                self._initial_state.meta.workflow_definition = self.workflow.__class__
            else:
                self._initial_state = self.workflow.get_state_at_node(node)
            self._entrypoints = entrypoint_nodes
        elif external_inputs:
            self._initial_state = self.workflow.get_most_recent_state()
            for descriptor, value in external_inputs.items():
                if not any(isinstance(value, type_) for type_ in descriptor.types):
                    raise NodeException(
                        f"Invalid external input type for {descriptor.name}",
                        code=WorkflowErrorCode.INVALID_INPUTS,
                    )
                self._initial_state.meta.external_inputs[descriptor] = value

            self._entrypoints = [
                ei.inputs_class.__parent_class__
                for ei in external_inputs
                if issubclass(ei.inputs_class.__parent_class__, BaseNode)
            ]
            self._is_resuming = True
        else:
            normalized_inputs = deepcopy(inputs) if inputs else self.workflow.get_default_inputs()
            if state:
                self._initial_state = deepcopy(state)
                self._initial_state.meta.workflow_inputs = normalized_inputs
                self._initial_state.meta.span_id = uuid4()
                self._initial_state.meta.workflow_definition = self.workflow.__class__
            else:
                self._initial_state = self.workflow.get_default_state(normalized_inputs)
                # We don't want to emit the initial state on the base case of Workflow Runs, since
                # all of that data is redundant and is derivable. It also clearly communicates that
                # there was no initial state provided by the user to invoke the workflow.
                self._should_emit_initial_state = False
            self._entrypoints = self.workflow.get_entrypoints()

        # This queue is responsible for sending events from WorkflowRunner to the outside world
        self._workflow_event_outer_queue: Queue[WorkflowEvent] = Queue()

        # This queue is responsible for sending events from the inner worker threads to WorkflowRunner
        self._workflow_event_inner_queue: Queue[WorkflowEvent] = Queue()

        self._max_concurrency = max_concurrency
        self._concurrency_queue: Queue[Tuple[StateType, Type[BaseNode], Optional[UUID]]] = Queue()

        # This queue is responsible for sending events from WorkflowRunner to the background thread
        # for user defined emitters
        self._background_thread_queue: Queue[BackgroundThreadItem] = Queue()

        self._dependencies: Dict[Type[BaseNode], Set[Type[BaseNode]]] = defaultdict(set)
        self._state_forks: Set[StateType] = {self._initial_state}

        self._active_nodes_by_execution_id: Dict[UUID, ActiveNode[StateType]] = {}
        self._cancel_signal = cancel_signal
        self._execution_context = init_execution_context or get_execution_context()

        setattr(
            self._initial_state,
            "__snapshot_callback__",
            lambda s, d: self._snapshot_state(s, d),
        )
        self.workflow.context._register_event_queue(self._workflow_event_inner_queue)
        self.workflow.context._register_node_output_mocks(node_output_mocks or [])

        self._outputs_listening_to_state = [
            descriptor for descriptor in self.workflow.Outputs if isinstance(descriptor.instance, StateValueReference)
        ]

    def _snapshot_state(self, state: StateType, deltas: List[StateDelta]) -> StateType:
        self._workflow_event_inner_queue.put(
            WorkflowExecutionSnapshottedEvent(
                trace_id=self._execution_context.trace_id,
                span_id=state.meta.span_id,
                body=WorkflowExecutionSnapshottedBody(
                    workflow_definition=self.workflow.__class__,
                    state=state,
                ),
                parent=self._execution_context.parent_context,
            )
        )

        delta_names = [d.name for d in deltas]
        for descriptor in self._outputs_listening_to_state:
            if not isinstance(descriptor.instance, StateValueReference):
                continue

            if descriptor.instance.name not in delta_names:
                continue

            resolved_delta = descriptor.instance.resolve(state)
            if resolved_delta is undefined:
                continue

            self._workflow_event_outer_queue.put(
                self._stream_workflow_event(
                    BaseOutput(
                        name=descriptor.name,
                        # We may want to switch this to using the delta from the state delta
                        # instead of the state value. In the short term, we need to show
                        # the full conversation of the tool calling node's delta's.
                        delta=resolved_delta,
                    )
                )
            )

        self.workflow._store.append_state_snapshot(state)
        self._background_thread_queue.put(state)
        return state

    def _emit_event(self, event: WorkflowEvent) -> WorkflowEvent:
        self.workflow._store.append_event(event)
        self._background_thread_queue.put(event)
        return event

    def _run_work_item(self, node: BaseNode[StateType], span_id: UUID) -> None:
        execution = get_execution_context()
        self._workflow_event_inner_queue.put(
            NodeExecutionInitiatedEvent(
                trace_id=execution.trace_id,
                span_id=span_id,
                body=NodeExecutionInitiatedBody(
                    node_definition=node.__class__,
                    inputs=node._inputs,
                ),
                parent=execution.parent_context,
            )
        )

        logger.debug(f"Started running node: {node.__class__.__name__}")

        try:
            updated_parent_context = NodeParentContext(
                span_id=span_id,
                node_definition=node.__class__,
                parent=execution.parent_context,
            )
            node_run_response: NodeRunResponse
            was_mocked: Optional[bool] = None
            mock_candidates = self.workflow.context.node_output_mocks_map.get(node.Outputs) or []
            for mock_candidate in mock_candidates:
                if mock_candidate.when_condition.resolve(node.state):
                    node_run_response = mock_candidate.then_outputs
                    was_mocked = True
                    break

            if not was_mocked:
                with execution_context(parent_context=updated_parent_context, trace_id=execution.trace_id):
                    node_run_response = node.run()

            ports = node.Ports()
            if not isinstance(node_run_response, (BaseOutputs, Iterator)):
                raise NodeException(
                    message=f"Node {node.__class__.__name__} did not return a valid node run response",
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                )

            if isinstance(node_run_response, BaseOutputs):
                if not isinstance(node_run_response, node.Outputs):
                    raise NodeException(
                        message=f"Node {node.__class__.__name__} did not return a valid outputs object",
                        code=WorkflowErrorCode.INVALID_OUTPUTS,
                    )

                outputs = node_run_response
            else:
                streaming_output_queues: Dict[str, Queue] = {}
                outputs = node.Outputs()

                def initiate_node_streaming_output(output: BaseOutput) -> None:
                    execution = get_execution_context()
                    streaming_output_queues[output.name] = Queue()
                    output_descriptor = OutputReference(
                        name=output.name,
                        types=(type(output.delta),),
                        instance=None,
                        outputs_class=node.Outputs,
                    )
                    with node.state.__quiet__():
                        node.state.meta.node_outputs[output_descriptor] = streaming_output_queues[output.name]
                    initiated_output: BaseOutput = BaseOutput(name=output.name)
                    initiated_ports = initiated_output > ports
                    self._workflow_event_inner_queue.put(
                        NodeExecutionStreamingEvent(
                            trace_id=execution.trace_id,
                            span_id=span_id,
                            body=NodeExecutionStreamingBody(
                                node_definition=node.__class__,
                                output=initiated_output,
                                invoked_ports=initiated_ports,
                            ),
                            parent=execution.parent_context,
                        ),
                    )

                with execution_context(parent_context=updated_parent_context, trace_id=execution.trace_id):
                    for output in node_run_response:
                        invoked_ports = output > ports
                        if output.is_initiated:
                            initiate_node_streaming_output(output)
                        elif output.is_streaming:
                            if output.name not in streaming_output_queues:
                                initiate_node_streaming_output(output)

                            streaming_output_queues[output.name].put(output.delta)
                            self._workflow_event_inner_queue.put(
                                NodeExecutionStreamingEvent(
                                    trace_id=execution.trace_id,
                                    span_id=span_id,
                                    body=NodeExecutionStreamingBody(
                                        node_definition=node.__class__,
                                        output=output,
                                        invoked_ports=invoked_ports,
                                    ),
                                    parent=execution.parent_context,
                                ),
                            )
                        elif output.is_fulfilled:
                            if output.name in streaming_output_queues:
                                streaming_output_queues[output.name].put(undefined)

                            setattr(outputs, output.name, output.value)
                            self._workflow_event_inner_queue.put(
                                NodeExecutionStreamingEvent(
                                    trace_id=execution.trace_id,
                                    span_id=span_id,
                                    body=NodeExecutionStreamingBody(
                                        node_definition=node.__class__,
                                        output=output,
                                        invoked_ports=invoked_ports,
                                    ),
                                    parent=execution.parent_context,
                                )
                            )

            node.state.meta.node_execution_cache.fulfill_node_execution(node.__class__, span_id)

            with node.state.__atomic__():
                for descriptor, output_value in outputs:
                    if output_value is undefined:
                        if descriptor in node.state.meta.node_outputs:
                            del node.state.meta.node_outputs[descriptor]
                        continue

                    node.state.meta.node_outputs[descriptor] = output_value

            invoked_ports = ports(outputs, node.state)
            self._workflow_event_inner_queue.put(
                NodeExecutionFulfilledEvent(
                    trace_id=execution.trace_id,
                    span_id=span_id,
                    body=NodeExecutionFulfilledBody(
                        node_definition=node.__class__,
                        outputs=outputs,
                        invoked_ports=invoked_ports,
                        mocked=was_mocked,
                    ),
                    parent=execution.parent_context,
                )
            )
        except NodeException as e:
            logger.info(e)

            self._workflow_event_inner_queue.put(
                NodeExecutionRejectedEvent(
                    trace_id=execution.trace_id,
                    span_id=span_id,
                    body=NodeExecutionRejectedBody(
                        node_definition=node.__class__,
                        error=e.error,
                    ),
                    parent=execution.parent_context,
                )
            )
        except WorkflowInitializationException as e:
            logger.info(e)
            self._workflow_event_inner_queue.put(
                NodeExecutionRejectedEvent(
                    trace_id=execution.trace_id,
                    span_id=span_id,
                    body=NodeExecutionRejectedBody(
                        node_definition=node.__class__,
                        error=e.error,
                    ),
                    parent=execution.parent_context,
                )
            )

        except Exception as e:
            error_message = self._parse_error_message(e)
            if error_message is None:
                logger.exception(f"An unexpected error occurred while running node {node.__class__.__name__}")
                error_code = WorkflowErrorCode.INTERNAL_ERROR
                error_message = "Internal error"
            else:
                error_code = WorkflowErrorCode.NODE_EXECUTION

            self._workflow_event_inner_queue.put(
                NodeExecutionRejectedEvent(
                    trace_id=execution.trace_id,
                    span_id=span_id,
                    body=NodeExecutionRejectedBody(
                        node_definition=node.__class__,
                        error=WorkflowError(
                            message=error_message,
                            code=error_code,
                        ),
                    ),
                    parent=execution.parent_context,
                ),
            )

        logger.debug(f"Finished running node: {node.__class__.__name__}")

    def _parse_error_message(self, exception: Exception) -> Optional[str]:
        try:
            _, _, tb = sys.exc_info()
            if tb:
                tb_list = traceback.extract_tb(tb)
                if tb_list:
                    last_frame = tb_list[-1]
                    exception_module = next(
                        (
                            mod.__name__
                            for mod in sys.modules.values()
                            if hasattr(mod, "__file__") and mod.__file__ == last_frame.filename
                        ),
                        None,
                    )
                    if exception_module and not exception_module.startswith("vellum."):
                        return str(exception)
        except Exception:
            pass

        return None

    def _context_run_work_item(
        self,
        node: BaseNode[StateType],
        span_id: UUID,
        parent_context: ParentContext,
        trace_id: UUID,
    ) -> None:
        with execution_context(
            parent_context=parent_context,
            trace_id=trace_id,
        ):
            self._run_work_item(node, span_id)

    def _handle_invoked_ports(
        self,
        state: StateType,
        ports: Optional[Iterable[Port]],
        invoked_by: Optional[UUID],
    ) -> None:
        if not ports:
            return

        for port in ports:
            for edge in port.edges:
                if port.fork_state:
                    next_state = deepcopy(state)
                    self._state_forks.add(next_state)
                else:
                    next_state = state

                if self._max_concurrency:
                    self._concurrency_queue.put((next_state, edge.to_node, invoked_by))
                else:
                    self._run_node_if_ready(next_state, edge.to_node, invoked_by)

        if self._max_concurrency:
            num_nodes_to_run = self._max_concurrency - len(self._active_nodes_by_execution_id)
            for _ in range(num_nodes_to_run):
                if self._concurrency_queue.empty():
                    break

                next_state, node_class, invoked_by = self._concurrency_queue.get()
                self._run_node_if_ready(next_state, node_class, invoked_by)

    def _run_node_if_ready(
        self,
        state: StateType,
        node_class: Type[BaseNode],
        invoked_by: Optional[UUID] = None,
    ) -> None:
        with state.__lock__:
            for descriptor in node_class.ExternalInputs:
                if not isinstance(descriptor, ExternalInputReference):
                    continue

                if state.meta.external_inputs.get(descriptor, undefined) is undefined:
                    state.meta.external_inputs[descriptor] = undefined
                    return

            all_deps = self._dependencies[node_class]
            node_span_id = node_class.Trigger._queue_node_execution(state, all_deps, invoked_by)

            try:
                if not node_class.Trigger.should_initiate(state, all_deps, node_span_id):
                    return
            except NodeException as e:
                execution = get_execution_context()

                self._workflow_event_outer_queue.put(
                    WorkflowExecutionRejectedEvent(
                        trace_id=execution.trace_id,
                        span_id=node_span_id,
                        body=WorkflowExecutionRejectedBody(
                            workflow_definition=self.workflow.__class__,
                            error=e.error,
                        ),
                    )
                )
                raise e

            execution = get_execution_context()
            node = node_class(state=state, context=self.workflow.context)
            state.meta.node_execution_cache.initiate_node_execution(node_class, node_span_id)
            self._active_nodes_by_execution_id[node_span_id] = ActiveNode(node=node)

            worker_thread = Thread(
                target=self._context_run_work_item,
                kwargs={
                    "node": node,
                    "span_id": node_span_id,
                    "parent_context": execution.parent_context,
                    "trace_id": execution.trace_id,
                },
            )
            worker_thread.start()

    def _handle_work_item_event(self, event: WorkflowEvent) -> Optional[WorkflowError]:
        active_node = self._active_nodes_by_execution_id.get(event.span_id)
        if not active_node:
            return None

        node = active_node.node
        if event.name == "node.execution.rejected":
            self._active_nodes_by_execution_id.pop(event.span_id)
            return event.error

        if event.name == "node.execution.streaming":
            for workflow_output_descriptor in self.workflow.Outputs:
                if node.__directly_emit_workflow_output__(event, workflow_output_descriptor):
                    active_node.was_outputs_streamed = True
                    self._workflow_event_outer_queue.put(
                        self._stream_workflow_event(
                            BaseOutput(
                                name=workflow_output_descriptor.name,
                                value=event.output.value,
                                delta=event.output.delta,
                            )
                        )
                    )
                    return None

                node_output_descriptor = workflow_output_descriptor.instance
                if not isinstance(node_output_descriptor, OutputReference):
                    continue
                if node_output_descriptor.outputs_class != event.node_definition.Outputs:
                    continue
                if node_output_descriptor.name != event.output.name:
                    continue

                active_node.was_outputs_streamed = True
                self._workflow_event_outer_queue.put(
                    self._stream_workflow_event(
                        BaseOutput(
                            name=workflow_output_descriptor.name,
                            value=event.output.value,
                            delta=event.output.delta,
                        )
                    )
                )

            self._handle_invoked_ports(node.state, event.invoked_ports, event.span_id)

            return None

        if event.name == "node.execution.fulfilled":
            self._active_nodes_by_execution_id.pop(event.span_id)
            if not active_node.was_outputs_streamed:
                for event_node_output_descriptor, node_output_value in event.outputs:
                    for workflow_output_descriptor in self.workflow.Outputs:
                        node_output_descriptor = workflow_output_descriptor.instance
                        if not isinstance(node_output_descriptor, OutputReference):
                            continue
                        if node_output_descriptor.outputs_class != event.node_definition.Outputs:
                            continue
                        if node_output_descriptor.name != event_node_output_descriptor.name:
                            continue

                        self._workflow_event_outer_queue.put(
                            self._stream_workflow_event(
                                BaseOutput(
                                    name=workflow_output_descriptor.name,
                                    value=node_output_value,
                                )
                            )
                        )

            self._handle_invoked_ports(node.state, event.invoked_ports, event.span_id)

            return None

        return None

    def _initiate_workflow_event(self) -> WorkflowExecutionInitiatedEvent:
        return WorkflowExecutionInitiatedEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionInitiatedBody(
                workflow_definition=self.workflow.__class__,
                inputs=self._initial_state.meta.workflow_inputs,
                initial_state=deepcopy(self._initial_state) if self._should_emit_initial_state else None,
            ),
            parent=self._execution_context.parent_context,
        )

    def _stream_workflow_event(self, output: BaseOutput) -> WorkflowExecutionStreamingEvent:
        return WorkflowExecutionStreamingEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionStreamingBody(
                workflow_definition=self.workflow.__class__,
                output=output,
            ),
            parent=self._execution_context.parent_context,
        )

    def _fulfill_workflow_event(self, outputs: OutputsType) -> WorkflowExecutionFulfilledEvent:
        return WorkflowExecutionFulfilledEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionFulfilledBody(
                workflow_definition=self.workflow.__class__,
                outputs=outputs,
            ),
            parent=self._execution_context.parent_context,
        )

    def _reject_workflow_event(self, error: WorkflowError) -> WorkflowExecutionRejectedEvent:
        return WorkflowExecutionRejectedEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionRejectedBody(
                workflow_definition=self.workflow.__class__,
                error=error,
            ),
            parent=self._execution_context.parent_context,
        )

    def _resume_workflow_event(self) -> WorkflowExecutionResumedEvent:
        return WorkflowExecutionResumedEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionResumedBody(
                workflow_definition=self.workflow.__class__,
            ),
            parent=self._execution_context.parent_context,
        )

    def _pause_workflow_event(self, external_inputs: Iterable[ExternalInputReference]) -> WorkflowExecutionPausedEvent:
        return WorkflowExecutionPausedEvent(
            trace_id=self._execution_context.trace_id,
            span_id=self._initial_state.meta.span_id,
            body=WorkflowExecutionPausedBody(
                workflow_definition=self.workflow.__class__,
                external_inputs=external_inputs,
            ),
            parent=self._execution_context.parent_context,
        )

    def _stream(self) -> None:
        for edge in self.workflow.get_edges():
            self._dependencies[edge.to_node].add(edge.from_port.node_class)

        current_parent = WorkflowParentContext(
            span_id=self._initial_state.meta.span_id,
            workflow_definition=self.workflow.__class__,
            parent=self._execution_context.parent_context,
            type="WORKFLOW",
        )
        for node_cls in self._entrypoints:
            try:
                if not self._max_concurrency or len(self._active_nodes_by_execution_id) < self._max_concurrency:
                    with execution_context(parent_context=current_parent, trace_id=self._execution_context.trace_id):
                        self._run_node_if_ready(self._initial_state, node_cls)
                else:
                    self._concurrency_queue.put((self._initial_state, node_cls, None))
            except NodeException as e:
                self._workflow_event_outer_queue.put(self._reject_workflow_event(e.error))
                return
            except WorkflowInitializationException as e:
                self._workflow_event_outer_queue.put(self._reject_workflow_event(e.error))
                return
            except Exception:
                err_message = f"An unexpected error occurred while initializing node {node_cls.__name__}"
                logger.exception(err_message)
                self._workflow_event_outer_queue.put(
                    self._reject_workflow_event(
                        WorkflowError(code=WorkflowErrorCode.INTERNAL_ERROR, message=err_message),
                    )
                )
                return

        rejection_error: Optional[WorkflowError] = None

        while True:
            if not self._active_nodes_by_execution_id:
                break

            event = self._workflow_event_inner_queue.get()

            self._workflow_event_outer_queue.put(event)

            with execution_context(parent_context=current_parent, trace_id=self._execution_context.trace_id):
                rejection_error = self._handle_work_item_event(event)

            if rejection_error:
                break

        # Handle any remaining events
        try:
            while event := self._workflow_event_inner_queue.get_nowait():
                self._workflow_event_outer_queue.put(event)

                with execution_context(parent_context=current_parent, trace_id=self._execution_context.trace_id):
                    rejection_error = self._handle_work_item_event(event)

                if rejection_error:
                    break
        except Empty:
            pass

        final_state = self._state_forks.pop()
        for other_state in self._state_forks:
            final_state += other_state

        unresolved_external_inputs = {
            descriptor
            for descriptor, node_input_value in final_state.meta.external_inputs.items()
            if node_input_value is undefined
        }
        if unresolved_external_inputs:
            self._workflow_event_outer_queue.put(
                self._pause_workflow_event(unresolved_external_inputs),
            )
            return

        if rejection_error:
            self._workflow_event_outer_queue.put(self._reject_workflow_event(rejection_error))
            return

        fulfilled_outputs = self.workflow.Outputs()
        for descriptor, value in fulfilled_outputs:
            if isinstance(value, BaseDescriptor):
                setattr(fulfilled_outputs, descriptor.name, value.resolve(final_state))
            elif isinstance(descriptor.instance, BaseDescriptor):
                setattr(
                    fulfilled_outputs,
                    descriptor.name,
                    descriptor.instance.resolve(final_state),
                )

        self._workflow_event_outer_queue.put(self._fulfill_workflow_event(fulfilled_outputs))

    def _run_background_thread(self) -> None:
        state_class = self.workflow.get_state_class()
        while True:
            item = self._background_thread_queue.get()
            if item is None:
                break

            if isinstance(item, state_class):
                for emitter in self.workflow.emitters:
                    emitter.snapshot_state(item)
            elif isinstance(item, BaseEvent):
                for emitter in self.workflow.emitters:
                    emitter.emit_event(item)

    def _run_cancel_thread(self, kill_switch: ThreadingEvent) -> None:
        if not self._cancel_signal:
            return

        while not kill_switch.wait(timeout=0.1):
            if self._cancel_signal.is_set():
                self._workflow_event_outer_queue.put(
                    self._reject_workflow_event(
                        WorkflowError(
                            code=WorkflowErrorCode.WORKFLOW_CANCELLED,
                            message="Workflow run cancelled",
                        )
                    )
                )
                return

    def _is_terminal_event(self, event: WorkflowEvent) -> bool:
        if (
            event.name == "workflow.execution.fulfilled"
            or event.name == "workflow.execution.rejected"
            or event.name == "workflow.execution.paused"
        ):
            return event.workflow_definition == self.workflow.__class__
        return False

    def _generate_events(self) -> Generator[WorkflowEvent, None, None]:
        background_thread = Thread(
            target=self._run_background_thread,
            name=f"{self.workflow.__class__.__name__}.background_thread",
        )
        background_thread.start()

        cancel_thread_kill_switch = ThreadingEvent()
        if self._cancel_signal:
            cancel_thread = Thread(
                target=self._run_cancel_thread,
                name=f"{self.workflow.__class__.__name__}.cancel_thread",
                kwargs={"kill_switch": cancel_thread_kill_switch},
            )
            cancel_thread.start()

        event: WorkflowEvent
        if self._is_resuming:
            event = self._resume_workflow_event()
        else:
            event = self._initiate_workflow_event()

        yield self._emit_event(event)

        # The extra level of indirection prevents the runner from waiting on the caller to consume the event stream
        stream_thread = Thread(
            target=self._stream,
            name=f"{self.workflow.__class__.__name__}.stream_thread",
        )
        stream_thread.start()

        while stream_thread.is_alive():
            try:
                event = self._workflow_event_outer_queue.get(timeout=0.1)
            except Empty:
                continue

            yield self._emit_event(event)

            if self._is_terminal_event(event):
                break

        try:
            while event := self._workflow_event_outer_queue.get_nowait():
                yield self._emit_event(event)
        except Empty:
            pass

        if not self._is_terminal_event(event):
            yield self._reject_workflow_event(
                WorkflowError(
                    code=WorkflowErrorCode.INTERNAL_ERROR,
                    message="An unexpected error occurred while streaming Workflow events",
                )
            )

        self._background_thread_queue.put(None)
        cancel_thread_kill_switch.set()

    def stream(self) -> WorkflowEventStream:
        return WorkflowEventGenerator(self._generate_events(), self._initial_state.meta.span_id)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .instructor_vectorizer_config import InstructorVectorizerConfig


class HkunlpInstructorXlVectorizer(UniversalBaseModel):
    """
    Vectorizer for hkunlp/instructor-xl.
    """

    model_name: typing.Literal["hkunlp/instructor-xl"] = "hkunlp/instructor-xl"
    config: InstructorVectorizerConfig

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

from .test_suite_run_deployment_release_tag_exec_config_request import TestSuiteRunDeploymentReleaseTagExecConfigRequest
from .test_suite_run_external_exec_config_request import TestSuiteRunExternalExecConfigRequest
from .test_suite_run_prompt_sandbox_exec_config_request import TestSuiteRunPromptSandboxExecConfigRequest
from .test_suite_run_prompt_sandbox_history_item_exec_config_request import (
    TestSuiteRunPromptSandboxHistoryItemExecConfigRequest,
)
from .test_suite_run_workflow_release_tag_exec_config_request import TestSuiteRunWorkflowReleaseTagExecConfigRequest
from .test_suite_run_workflow_sandbox_exec_config_request import TestSuiteRunWorkflowSandboxExecConfigRequest
from .test_suite_run_workflow_sandbox_history_item_exec_config_request import (
    TestSuiteRunWorkflowSandboxHistoryItemExecConfigRequest,
)

TestSuiteRunExecConfigRequest = typing.Union[
    TestSuiteRunDeploymentReleaseTagExecConfigRequest,
    TestSuiteRunPromptSandboxExecConfigRequest,
    TestSuiteRunPromptSandboxHistoryItemExecConfigRequest,
    TestSuiteRunWorkflowReleaseTagExecConfigRequest,
    TestSuiteRunWorkflowSandboxExecConfigRequest,
    TestSuiteRunWorkflowSandboxHistoryItemExecConfigRequest,
    TestSuiteRunExternalExecConfigRequest,
]



# This file was auto-generated by Fern from our API Definition.

import typing

from .reducto_chunking import ReductoChunking
from .sentence_chunking import SentenceChunking
from .token_overlapping_window_chunking import TokenOverlappingWindowChunking

DocumentIndexChunking = typing.Union[ReductoChunking, SentenceChunking, TokenOverlappingWindowChunking]



# This file was auto-generated by Fern from our API Definition.

import typing

DocumentProcessingState = typing.Union[
    typing.Literal["QUEUED", "PROCESSING", "PROCESSED", "FAILED", "UNKNOWN"], typing.Any
]



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import WorkspaceSecretRead
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_api_node.workflow import SimpleAPIWorkflow


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow that uses a vellum API node
    # AND stubbed out API calls
    workspace_secret_id = str(uuid4())
    workspace_secret = WorkspaceSecretRead(
        id=workspace_secret_id,
        modified=datetime.now(),
        name="MY_SECRET",
        label="My Secret",
        secret_type="USER_DEFINED",
    )
    vellum_client.workspace_secrets.retrieve.return_value = workspace_secret

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=SimpleAPIWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 3
    assert not DeepDiff(
        [
            {"id": "9a37bf7d-484e-4725-903e-f3254df38a0a", "key": "json", "type": "JSON"},
            {"id": "5090e96d-5787-4a08-bf58-129101cf2548", "key": "headers", "type": "JSON"},
            {"id": "44ea8d75-e2a8-4627-85b1-8504b65d25c9", "key": "status_code", "type": "NUMBER"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 4
    assert len(workflow_raw_data["nodes"]) == 5

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "3a6b1467-5c83-4bcd-86a0-6415bc32d23b",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "8eaa7f02-25ff-4a00-9b0a-5185718d89b3"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    api_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
            "type": "API",
            "inputs": [
                {
                    "id": "cd8a19f4-4eb7-4359-a8a6-918569c466a5",
                    "key": "url",
                    "value": {
                        "rules": [
                            {"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "https://api.vellum.ai"}}
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "fd61b5ac-39f9-4cfe-a839-f8ce78c202df",
                    "key": "method",
                    "value": {
                        "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "POST"}}],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "c3a17ceb-e201-4025-b18a-9162aac7705e",
                    "key": "body",
                    "value": {
                        "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": {"key": "value"}}}],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "3092bf23-3202-4f3e-874c-9a33ccc73459",
                    "key": "authorization_type",
                    "value": {
                        "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "API_KEY"}}],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "fee5e3c9-442a-4922-ba80-5ee07361cea7",
                    "key": "bearer_token_value",
                    "value": {
                        "rules": [
                            {"type": "WORKSPACE_SECRET", "data": {"type": "STRING", "workspace_secret_id": None}}
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "2fcdfbc3-8095-4277-bb4a-a201fd326b54",
                    "key": "api_key_header_key",
                    "value": {
                        "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "CUSTOM_API_KEY"}}],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "d794bb51-a419-4fd8-be63-dfaf4166e831",
                    "key": "api_key_header_value",
                    "value": {
                        "rules": [
                            {
                                "type": "WORKSPACE_SECRET",
                                "data": {
                                    "type": "STRING",
                                    "workspace_secret_id": f"{workspace_secret_id}",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "57c31247-998a-430d-bb62-bf50eca7df35",
                    "key": "additional_header_key",
                    "value": {
                        "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "additional_header"}}],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "47b32274-f19b-4c15-b788-55c069c311c5",
                    "key": "additional_header_value",
                    "value": {
                        "rules": [
                            {"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "additional header value"}}
                        ],
                        "combinator": "OR",
                    },
                },
            ],
            "attributes": [
                {
                    "id": "ad719e65-0032-4012-a0bd-9b5162194bce",
                    "name": "timeout",
                    "value": {
                        "type": "CONSTANT_VALUE",
                        "value": {"type": "JSON", "value": None},
                    },
                },
            ],
            "data": {
                "label": "Simple A P I Node",
                "error_output_id": None,
                "source_handle_id": "7c33b4d3-9204-4bd5-9371-80ee34f83073",
                "target_handle_id": "14b538a5-aedb-41f3-b579-039956b7c7ed",
                "url_input_id": "cd8a19f4-4eb7-4359-a8a6-918569c466a5",
                "method_input_id": "fd61b5ac-39f9-4cfe-a839-f8ce78c202df",
                "body_input_id": "c3a17ceb-e201-4025-b18a-9162aac7705e",
                "authorization_type_input_id": "3092bf23-3202-4f3e-874c-9a33ccc73459",
                "bearer_token_value_input_id": "fee5e3c9-442a-4922-ba80-5ee07361cea7",
                "api_key_header_key_input_id": "2fcdfbc3-8095-4277-bb4a-a201fd326b54",
                "api_key_header_value_input_id": "d794bb51-a419-4fd8-be63-dfaf4166e831",
                "additional_headers": [
                    {
                        "header_key_input_id": "57c31247-998a-430d-bb62-bf50eca7df35",
                        "header_value_input_id": "47b32274-f19b-4c15-b788-55c069c311c5",
                    }
                ],
                "text_output_id": "17342c21-12bb-49ab-88ce-f144e0376b32",
                "json_output_id": "12e4a99d-883d-4da5-aa51-35817d94013e",
                "status_code_output_id": "fecc16c3-400e-4fd3-8223-08366070e3b1",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "APINode",
                "module": ["vellum", "workflows", "nodes", "displayable", "api_node", "node"],
            },
            "definition": {
                "name": "SimpleAPINode",
                "module": ["tests", "workflows", "basic_api_node", "workflow"],
            },
            "trigger": {
                "id": "14b538a5-aedb-41f3-b579-039956b7c7ed",
                "merge_behavior": "AWAIT_ANY",
            },
            "ports": [{"id": "7c33b4d3-9204-4bd5-9371-80ee34f83073", "name": "default", "type": "DEFAULT"}],
        },
        api_node,
    )

    final_output_nodes = workflow_raw_data["nodes"][2:5]
    assert not DeepDiff(
        [
            {
                "id": "8f975ab1-aca6-4dc1-aa80-c596f4e13afa",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "json",
                    "target_handle_id": "06853542-e1a1-4a00-bd1e-4ac40f347b32",
                    "output_id": "9a37bf7d-484e-4725-903e-f3254df38a0a",
                    "output_type": "JSON",
                    "node_input_id": "b49f0f85-37fc-4686-81a7-287c06634661",
                },
                "inputs": [
                    {
                        "id": "b49f0f85-37fc-4686-81a7-287c06634661",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                                        "output_id": "12e4a99d-883d-4da5-aa51-35817d94013e",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 400.0, "y": 200.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "736f9bd0-f487-42af-bdb3-780b4941c61c",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "headers",
                    "target_handle_id": "80d0894f-642e-4d2e-b43a-f236e7bedb3c",
                    "output_id": "5090e96d-5787-4a08-bf58-129101cf2548",
                    "output_type": "JSON",
                    "node_input_id": "5e892e5b-0004-4a04-bd2e-1ea9e0e5d3f9",
                },
                "inputs": [
                    {
                        "id": "5e892e5b-0004-4a04-bd2e-1ea9e0e5d3f9",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                                        "output_id": "0d76e1e1-3a4b-4eb4-a606-f73d62cf1a7e",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 400.0, "y": -50.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "3f3ffc50-b156-48ac-b5f3-f68cb05c2b90",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "status_code",
                    "target_handle_id": "0c98c306-b519-40d7-8b05-321b1dfd7f11",
                    "output_id": "44ea8d75-e2a8-4627-85b1-8504b65d25c9",
                    "output_type": "NUMBER",
                    "node_input_id": "14345321-7e6b-4e2a-918a-7a5b0064f047",
                },
                "inputs": [
                    {
                        "id": "14345321-7e6b-4e2a-918a-7a5b0064f047",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                                        "output_id": "fecc16c3-400e-4fd3-8223-08366070e3b1",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 400.0, "y": -300.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
        ],
        final_output_nodes,
        ignore_order=True,
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "3c7551f5-0a76-4274-953e-c048b15f560a",
                "source_node_id": "3a6b1467-5c83-4bcd-86a0-6415bc32d23b",
                "source_handle_id": "8eaa7f02-25ff-4a00-9b0a-5185718d89b3",
                "target_node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                "target_handle_id": "14b538a5-aedb-41f3-b579-039956b7c7ed",
                "type": "DEFAULT",
            },
            {
                "id": "422bbbb7-38b2-4e19-ac95-24a86ed24100",
                "source_node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                "source_handle_id": "7c33b4d3-9204-4bd5-9371-80ee34f83073",
                "target_node_id": "8f975ab1-aca6-4dc1-aa80-c596f4e13afa",
                "target_handle_id": "06853542-e1a1-4a00-bd1e-4ac40f347b32",
                "type": "DEFAULT",
            },
            {
                "id": "73007fe0-5cc2-435e-b2e4-6fa734153fbd",
                "source_node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                "source_handle_id": "7c33b4d3-9204-4bd5-9371-80ee34f83073",
                "target_node_id": "736f9bd0-f487-42af-bdb3-780b4941c61c",
                "target_handle_id": "80d0894f-642e-4d2e-b43a-f236e7bedb3c",
                "type": "DEFAULT",
            },
            {
                "id": "4ac7d1c4-697b-440c-a35d-61ffb44a33b3",
                "source_node_id": "facb80d7-ee08-42a0-82a9-ee26a9218185",
                "source_handle_id": "7c33b4d3-9204-4bd5-9371-80ee34f83073",
                "target_node_id": "3f3ffc50-b156-48ac-b5f3-f68cb05c2b90",
                "target_handle_id": "0c98c306-b519-40d7-8b05-321b1dfd7f11",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "SimpleAPIWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_api_node",
            "workflow",
        ],
    }



# This file was auto-generated by Fern from our API Definition.

import typing

LogicalOperator = typing.Union[
    typing.Literal[
        "=",
        "!=",
        "<",
        ">",
        "<=",
        ">=",
        "contains",
        "beginsWith",
        "endsWith",
        "doesNotContain",
        "doesNotBeginWith",
        "doesNotEndWith",
        "null",
        "notNull",
        "in",
        "notIn",
        "between",
        "notBetween",
        "blank",
        "notBlank",
        "coalesce",
        "accessField",
        "parseJson",
        "and",
        "or",
        "isError",
    ],
    typing.Any,
]
